# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import enum
import random
import string
from typing import Collection, Dict, List, Mapping, Optional, Sequence, Set, Type, TypeVar

from gt4py.eve.concepts import (
    AnySourceLocation,
    FrozenNode,
    Node,
    SourceLocation,
    SourceLocationGroup,
    SymbolName,
    VType,
)
from gt4py.eve.datamodels import Coerced
from gt4py.eve.traits import SymbolTableTrait, ValidatedSymbolTableTrait
from gt4py.eve.type_definitions import IntEnum, StrEnum
from gt4py.eve.utils import CaseStyleConverter


T = TypeVar("T")
S = TypeVar("S")


# -- Node definitions --
@enum.unique
class IntKind(IntEnum):
    """Sample int Enum."""

    MINUS = -1
    ZERO = 0
    PLUS = 1


@enum.unique
class StrKind(StrEnum):
    """Sample string Enum."""

    FOO = "foo"
    BLA = "bla"
    FIZ = "fiz"
    FUZ = "fuz"


SimpleVType = VType("simple")


class EmptyNode(Node):
    pass


class LocationNode(Node):
    loc: SourceLocation


class SimpleNode(Node):
    int_value: int
    bool_value: bool
    float_value: float
    str_value: str
    bytes_value: bytes
    int_kind: IntKind
    str_kind: StrKind


class SimpleNodeWithOptionals(Node):
    int_value: int
    float_value: Optional[float]
    str_value: Optional[str] = None


class SimpleNodeWithLoc(Node):
    int_value: int
    float_value: float
    str_value: str
    loc: Optional[AnySourceLocation]


class SimpleNodeWithCollections(Node):
    int_value: int
    int_list: List[int]
    str_set: Set[str]
    str_to_int_dict: Dict[str, int]
    loc: Optional[AnySourceLocation]


class SimpleNodeWithAbstractCollections(Node):
    int_value: int
    int_sequence: Sequence[int]
    str_set: Set[str]
    str_to_int_mapping: Mapping[str, int]
    loc: Optional[AnySourceLocation] = None


class SimpleNodeWithSymbolName(Node):
    int_value: int
    name: Coerced[SymbolName]


class SimpleNodeWithDefaultSymbolName(Node):
    int_value: int
    name: SymbolName = SymbolName("symbol_name")


class CompoundNode(Node):
    int_value: int
    location: LocationNode
    simple: SimpleNode
    simple_loc: SimpleNodeWithLoc
    simple_opt: SimpleNodeWithOptionals
    other_simple_opt: Optional[SimpleNodeWithOptionals]


class CompoundNodeWithSymbols(Node):
    int_value: int
    location: LocationNode
    simple: SimpleNode
    simple_loc: SimpleNodeWithLoc
    simple_opt: SimpleNodeWithOptionals
    other_simple_opt: Optional[SimpleNodeWithOptionals]
    node_with_name: SimpleNodeWithSymbolName


class NodeWithSymbolTable(Node, SymbolTableTrait):
    node_with_name: SimpleNodeWithSymbolName
    list_with_name: List[SimpleNodeWithSymbolName]
    node_with_default_name: SimpleNodeWithDefaultSymbolName
    compound_with_name: CompoundNodeWithSymbols


class NodeWithValidatedSymbolTable(NodeWithSymbolTable, ValidatedSymbolTableTrait):
    node_with_name: SimpleNodeWithSymbolName
    list_with_name: List[SimpleNodeWithSymbolName]
    node_with_default_name: SimpleNodeWithDefaultSymbolName
    compound_with_name: CompoundNodeWithSymbols


class FrozenSimpleNode(FrozenNode):
    int_value: int
    bool_value: bool
    float_value: float
    str_value: str
    bytes_value: bytes
    int_kind: IntKind
    str_kind: StrKind


# -- General maker functions --
_MIN_INT = -9999
_MAX_INT = 9999
_MIN_FLOAT = -999.0
_MAX_FLOAT = 999.09
_SEQUENCE_LEN = 6


def make_random_bool_value() -> bool:
    return random.choice([True, False])


def make_bool_value(*, fixed: bool = False) -> bool:
    return True if fixed else make_random_bool_value()


def make_random_int_value() -> int:
    return random.randint(_MIN_INT, _MAX_INT)


def make_int_value(*, fixed: bool = False) -> int:
    return 1 if fixed else make_random_int_value()


def make_random_neg_int_value() -> int:
    return random.randint(_MIN_INT, 1)


def make_neg_int_value(*, fixed: bool = False) -> int:
    return -2 if fixed else make_random_neg_int_value()


def make_random_pos_int_value() -> int:
    return random.randint(1, _MAX_INT)


def make_pos_int_value(*, fixed: bool = False) -> int:
    return 2 if fixed else make_random_pos_int_value()


def make_random_float_value() -> float:
    return _MIN_FLOAT + random.random() * (_MAX_FLOAT - _MIN_FLOAT)


def make_float_value(*, fixed: bool = False) -> float:
    return 1.1 if fixed else make_random_float_value()


def make_random_str_value(length: Optional[int] = None) -> str:
    length = length or _SEQUENCE_LEN
    return "".join(random.choice(string.ascii_letters) for _ in range(length))


def make_str_value(length: int = _SEQUENCE_LEN, *, fixed: bool = False) -> str:
    return string.ascii_letters[:length] if fixed else make_random_str_value(length)


def make_random_member_value(values: Sequence[T]) -> T:
    return random.choice(values)


def make_member_value(values: Sequence[T], *, fixed: bool = False) -> T:
    return values[0] if fixed else make_random_member_value(values)


def make_collection_value(
    item_type: Type[T],
    collection_type: Type[Collection[T]] = list,
    length: Optional[int] = None,
    *,
    fixed: bool = False,
) -> Collection[T]:
    length = length or _SEQUENCE_LEN

    maker_attr_name = f"make_{item_type.__name__}"
    try:
        maker = globals()[maker_attr_name]
    except Exception:

        def maker():
            return item_type()

    return collection_type([maker() for _ in range(length)])  # type: ignore


def make_mapping_value(
    key_type: Type[S],
    value_type: Type[T],
    mapping_type: Type[Mapping[S, T]] = dict,
    length: Optional[int] = None,
    *,
    fixed: bool = False,
) -> Mapping[S, T]:
    length = length or _SEQUENCE_LEN

    key_maker_attr_name = f"make_{key_type.__name__}"
    try:
        key_maker = globals()[key_maker_attr_name]
    except Exception:

        def key_maker():
            return key_type()

    value_maker_attr_name = f"make_{value_type.__name__}"
    try:
        value_maker = globals()[value_maker_attr_name]
    except Exception:

        def value_maker():
            return value_type()

    return mapping_type({key_maker(): value_maker() for _ in range(length)})  # type: ignore


def make_multinode_collection_value(
    node_class: Type[Node],
    collection_type: Type[Collection[T]] = list,
    length: Optional[int] = None,
    *,
    fixed: bool = False,
) -> Collection[T]:
    length = length or _SEQUENCE_LEN

    item_maker = globals()[
        f"make_{CaseStyleConverter.convert(node_class.__name__, 'pascal', 'snake')}"
    ]
    items = [item_maker(fixed=fixed) for _ in range(length)]

    return collection_type(items)  # type: ignore


# -- Node maker functions --
def make_source_location(*, fixed: bool = False) -> SourceLocation:
    line = make_pos_int_value(fixed=fixed)
    column = make_pos_int_value(fixed=fixed)
    str_value = make_str_value(fixed=fixed)
    source = f"file_{str_value}.py"

    return SourceLocation(line=line, column=column, source=source)


def make_source_location_group(*, fixed: bool = False) -> SourceLocationGroup:
    loc1 = make_source_location(fixed=fixed)
    loc2 = make_source_location(fixed=fixed)

    return SourceLocationGroup(loc1, loc2, context=make_str_value(fixed=fixed))


def make_empty_node(*, fixed: bool = False) -> EmptyNode:
    return EmptyNode()


def make_location_node(*, fixed: bool = False) -> LocationNode:
    return LocationNode(loc=make_source_location(fixed=fixed))


def make_simple_node(*, fixed: bool = False) -> SimpleNode:
    int_value = make_int_value(fixed=fixed)
    bool_value = make_bool_value(fixed=fixed)
    float_value = make_float_value(fixed=fixed)
    str_value = make_str_value(fixed=fixed)
    bytes_value = make_str_value(fixed=fixed).encode()
    int_kind = IntKind.PLUS if fixed else make_member_value([*IntKind], fixed=fixed)
    str_kind = StrKind.BLA if fixed else make_member_value([*StrKind], fixed=fixed)

    return SimpleNode(
        int_value=int_value,
        bool_value=bool_value,
        float_value=float_value,
        str_value=str_value,
        bytes_value=bytes_value,
        int_kind=int_kind,
        str_kind=str_kind,
    )


def make_simple_node_with_optionals(*, fixed: bool = False) -> SimpleNodeWithOptionals:
    int_value = make_int_value(fixed=fixed)
    float_value = make_float_value(fixed=fixed)

    return SimpleNodeWithOptionals(int_value=int_value, float_value=float_value)


def make_simple_node_with_loc(*, fixed: bool = False) -> SimpleNodeWithLoc:
    int_value = make_int_value(fixed=fixed)
    float_value = make_float_value(fixed=fixed)
    str_value = make_str_value(fixed=fixed)
    loc = make_source_location(fixed=fixed)

    return SimpleNodeWithLoc(
        int_value=int_value, float_value=float_value, str_value=str_value, loc=loc
    )


def make_simple_node_with_collections(*, fixed: bool = False) -> SimpleNodeWithCollections:
    int_value = make_int_value(fixed=fixed)
    int_list = make_collection_value(int, length=3)
    str_set = make_collection_value(str, set, length=3)
    str_to_int_dict = make_mapping_value(key_type=str, value_type=int, length=3)
    loc = make_source_location(fixed=fixed)

    return SimpleNodeWithCollections(
        int_value=int_value,
        int_list=int_list,
        str_set=str_set,
        str_to_int_dict=str_to_int_dict,
        loc=loc,
    )


def make_simple_node_with_abstract_collections(
    *,
    fixed: bool = False,
) -> SimpleNodeWithAbstractCollections:
    int_value = make_int_value(fixed=fixed)
    int_sequence = make_collection_value(int, collection_type=tuple, length=3)
    str_set = make_collection_value(str, set, length=3)
    str_to_int_mapping = make_mapping_value(key_type=str, value_type=int, length=3)

    return SimpleNodeWithAbstractCollections(
        int_value=int_value,
        int_sequence=int_sequence,
        str_set=str_set,
        str_to_int_mapping=str_to_int_mapping,
    )


def make_simple_node_with_symbol_name(
    *,
    fixed: bool = False,
) -> SimpleNodeWithSymbolName:
    int_value = make_int_value(fixed=fixed)
    name = make_str_value(fixed=fixed)

    return SimpleNodeWithSymbolName(int_value=int_value, name=name)


def make_simple_node_with_default_symbol_name(
    *,
    fixed: bool = False,
) -> SimpleNodeWithDefaultSymbolName:
    int_value = make_int_value(fixed=fixed)

    return SimpleNodeWithDefaultSymbolName(int_value=int_value)


def make_compound_node(*, fixed: bool = False) -> CompoundNode:
    return CompoundNode(
        int_value=make_int_value(fixed=fixed),
        location=make_location_node(),
        simple=make_simple_node(),
        simple_loc=make_simple_node_with_loc(),
        simple_opt=make_simple_node_with_optionals(),
        other_simple_opt=None,
    )


def make_compound_node_with_symbols(*, fixed: bool = False) -> CompoundNodeWithSymbols:
    return CompoundNodeWithSymbols(
        int_value=make_int_value(fixed=fixed),
        location=make_location_node(),
        simple=make_simple_node(),
        simple_loc=make_simple_node_with_loc(),
        simple_opt=make_simple_node_with_optionals(),
        other_simple_opt=None,
        node_with_name=make_simple_node_with_symbol_name(fixed=fixed),
    )


def make_node_with_symbol_table(*, fixed: bool = False) -> NodeWithSymbolTable:
    return NodeWithSymbolTable(
        node_with_name=make_simple_node_with_symbol_name(fixed=fixed),
        node_with_default_name=make_simple_node_with_default_symbol_name(fixed=fixed),
        list_with_name=make_multinode_collection_value(
            SimpleNodeWithSymbolName, length=4, fixed=fixed
        ),
        compound_with_name=make_compound_node_with_symbols(fixed=fixed),
    )


def make_node_with_validate_symbol_table(*, fixed: bool = False) -> NodeWithValidatedSymbolTable:
    return NodeWithValidatedSymbolTable(
        node_with_name=make_simple_node_with_symbol_name(fixed=fixed),
        node_with_default_name=make_simple_node_with_default_symbol_name(fixed=fixed),
        list_with_name=make_multinode_collection_value(
            SimpleNodeWithSymbolName, length=4, fixed=fixed
        ),
        compound_with_name=make_compound_node_with_symbols(fixed=fixed),
    )


def make_frozen_simple_node(*, fixed: bool = False) -> FrozenSimpleNode:
    int_value = make_int_value(fixed=fixed)
    bool_value = make_bool_value(fixed=fixed)
    float_value = make_float_value(fixed=fixed)
    str_value = make_str_value(fixed=fixed)
    bytes_value = make_str_value(fixed=fixed).encode()
    int_kind = IntKind.PLUS if fixed else make_member_value([*IntKind], fixed=fixed)
    str_kind = StrKind.BLA if fixed else make_member_value([*StrKind], fixed=fixed)

    return FrozenSimpleNode(
        int_value=int_value,
        bool_value=bool_value,
        float_value=float_value,
        str_value=str_value,
        bytes_value=bytes_value,
        int_kind=int_kind,
        str_kind=str_kind,
    )


# -- Makers of invalid nodes --
def make_invalid_location_node(*, fixed: bool = False) -> LocationNode:
    return LocationNode(loc=SourceLocation(line=0, column=-1, source="<str>"))


def make_invalid_at_int_simple_node(*, fixed: bool = False) -> SimpleNode:
    int_value = make_float_value(fixed=fixed)
    bool_value = make_bool_value(fixed=fixed)
    float_value = make_float_value(fixed=fixed)
    bytes_value = make_str_value(fixed=fixed).encode()
    str_value = make_str_value(fixed=fixed)
    int_kind = IntKind.PLUS if fixed else make_member_value([*IntKind], fixed=fixed)
    str_kind = StrKind.BLA if fixed else make_member_value([*StrKind], fixed=fixed)

    return SimpleNode(
        int_value=int_value,
        bool_value=bool_value,
        float_value=float_value,
        str_value=str_value,
        bytes_value=bytes_value,
        int_kind=int_kind,
        str_kind=str_kind,
    )


def make_invalid_at_float_simple_node(*, fixed: bool = False) -> SimpleNode:
    int_value = make_int_value(fixed=fixed)
    bool_value = make_bool_value(fixed=fixed)
    float_value = make_int_value(fixed=fixed)
    str_value = make_str_value(fixed=fixed)
    bytes_value = make_str_value(fixed=fixed).encode()
    int_kind = IntKind.PLUS if fixed else make_member_value([*IntKind], fixed=fixed)
    str_kind = StrKind.BLA if fixed else make_member_value([*StrKind], fixed=fixed)

    return SimpleNode(
        int_value=int_value,
        bool_value=bool_value,
        float_value=float_value,
        str_value=str_value,
        bytes_value=bytes_value,
        int_kind=int_kind,
        str_kind=str_kind,
    )


def make_invalid_at_str_simple_node(*, fixed: bool = False) -> SimpleNode:
    int_value = make_int_value(fixed=fixed)
    bool_value = make_bool_value(fixed=fixed)
    float_value = make_float_value(fixed=fixed)
    str_value = make_float_value(fixed=fixed)
    bytes_value = make_str_value(fixed=fixed).encode()
    int_kind = IntKind.PLUS if fixed else make_member_value([*IntKind])
    str_kind = StrKind.BLA if fixed else make_member_value([*StrKind])

    return SimpleNode(
        int_value=int_value,
        bool_value=bool_value,
        float_value=float_value,
        str_value=str_value,
        bytes_value=bytes_value,
        int_kind=int_kind,
        str_kind=str_kind,
    )


def make_invalid_at_bytes_simple_node(*, fixed: bool = False) -> SimpleNode:
    int_value = make_int_value(fixed=fixed)
    bool_value = make_bool_value(fixed=fixed)
    float_value = make_float_value(fixed=fixed)
    str_value = make_str_value(fixed=fixed)
    bytes_value = [1, "2", (3, 4)]
    int_kind = IntKind.PLUS if fixed else make_member_value([*IntKind])
    str_kind = StrKind.BLA if fixed else make_member_value([*StrKind])

    return SimpleNode(
        int_value=int_value,
        bool_value=bool_value,
        float_value=float_value,
        str_value=str_value,
        bytes_value=bytes_value,
        int_kind=int_kind,
        str_kind=str_kind,
    )


def make_invalid_at_enum_simple_node(*, fixed: bool = False) -> SimpleNode:
    int_value = make_int_value(fixed=fixed)
    bool_value = make_bool_value(fixed=fixed)
    float_value = make_float_value(fixed=fixed)
    str_value = make_str_value(fixed=fixed)
    bytes_value = make_str_value(fixed=fixed).encode()
    int_kind = "asdf"
    str_kind = StrKind.BLA if fixed else make_member_value([*StrKind])

    return SimpleNode(
        int_value=int_value,
        bool_value=bool_value,
        float_value=float_value,
        str_value=str_value,
        bytes_value=bytes_value,
        int_kind=int_kind,
        str_kind=str_kind,
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses
from dataclasses import dataclass
from typing import Any, ChainMap, Dict, List, Optional, Set, Tuple

import dace
import dace.data
import dace.library
import dace.subsets

from gt4py import eve
from gt4py.cartesian.gtc import daceir as dcir
from gt4py.cartesian.gtc.dace.expansion.tasklet_codegen import TaskletCodegen
from gt4py.cartesian.gtc.dace.expansion.utils import get_dace_debuginfo
from gt4py.cartesian.gtc.dace.symbol_utils import data_type_to_dace_typeclass
from gt4py.cartesian.gtc.dace.utils import make_dace_subset


class StencilComputationSDFGBuilder(eve.VisitorWithSymbolTableTrait):
    @dataclass
    class NodeContext:
        input_node_and_conns: Dict[Optional[str], Tuple[dace.nodes.Node, Optional[str]]]
        output_node_and_conns: Dict[Optional[str], Tuple[dace.nodes.Node, Optional[str]]]

    @dataclass
    class SDFGContext:
        sdfg: dace.SDFG
        state: dace.SDFGState
        state_stack: List[dace.SDFGState] = dataclasses.field(default_factory=list)

        def add_state(self):
            new_state = self.sdfg.add_state()
            for edge in self.sdfg.out_edges(self.state):
                self.sdfg.remove_edge(edge)
                self.sdfg.add_edge(new_state, edge.dst, edge.data)
            self.sdfg.add_edge(self.state, new_state, dace.InterstateEdge())
            self.state = new_state
            return self

        def add_loop(self, index_range: dcir.Range):
            loop_state = self.sdfg.add_state()
            after_state = self.sdfg.add_state()
            for edge in self.sdfg.out_edges(self.state):
                self.sdfg.remove_edge(edge)
                self.sdfg.add_edge(
                    after_state,
                    edge.dst,
                    edge.data,
                )

            assert isinstance(index_range.interval, dcir.DomainInterval)
            if index_range.stride < 0:
                initialize_expr = f"{index_range.interval.end} - 1"
                end_expr = f"{index_range.interval.start} - 1"
            else:
                initialize_expr = str(index_range.interval.start)
                end_expr = str(index_range.interval.end)
            comparison_op = "<" if index_range.stride > 0 else ">"
            condition_expr = f"{index_range.var} {comparison_op} {end_expr}"
            _, _, after_state = self.sdfg.add_loop(
                before_state=self.state,
                loop_state=loop_state,
                after_state=after_state,
                loop_var=index_range.var,
                initialize_expr=initialize_expr,
                condition_expr=condition_expr,
                increment_expr=f"{index_range.var}+({index_range.stride})",
            )
            if index_range.var not in self.sdfg.symbols:
                self.sdfg.add_symbol(index_range.var, stype=dace.int32)

            self.state_stack.append(after_state)
            self.state = loop_state
            return self

        def pop_loop(self):
            self.state = self.state_stack[-1]
            del self.state_stack[-1]

    def visit_Memlet(
        self,
        node: dcir.Memlet,
        *,
        scope_node: dcir.ComputationNode,
        sdfg_ctx: "StencilComputationSDFGBuilder.SDFGContext",
        node_ctx: "StencilComputationSDFGBuilder.NodeContext",
        connector_prefix="",
        symtable: ChainMap[eve.SymbolRef, dcir.Decl],
    ) -> None:
        field_decl = symtable[node.field]
        assert isinstance(field_decl, dcir.FieldDecl)
        memlet = dace.Memlet(
            node.field,
            subset=make_dace_subset(field_decl.access_info, node.access_info, field_decl.data_dims),
            dynamic=field_decl.is_dynamic,
        )
        if node.is_read:
            sdfg_ctx.state.add_edge(
                *node_ctx.input_node_and_conns[memlet.data],
                scope_node,
                connector_prefix + node.connector,
                memlet,
            )
        if node.is_write:
            sdfg_ctx.state.add_edge(
                scope_node,
                connector_prefix + node.connector,
                *node_ctx.output_node_and_conns[memlet.data],
                memlet,
            )

    @classmethod
    def _add_empty_edges(
        cls,
        entry_node: dace.nodes.Node,
        exit_node: dace.nodes.Node,
        *,
        sdfg_ctx: "StencilComputationSDFGBuilder.SDFGContext",
        node_ctx: "StencilComputationSDFGBuilder.NodeContext",
    ) -> None:
        if not sdfg_ctx.state.in_degree(entry_node) and None in node_ctx.input_node_and_conns:
            sdfg_ctx.state.add_edge(
                *node_ctx.input_node_and_conns[None], entry_node, None, dace.Memlet()
            )
        if not sdfg_ctx.state.out_degree(exit_node) and None in node_ctx.output_node_and_conns:
            sdfg_ctx.state.add_edge(
                exit_node, None, *node_ctx.output_node_and_conns[None], dace.Memlet()
            )

    def visit_Tasklet(
        self,
        node: dcir.Tasklet,
        *,
        sdfg_ctx: "StencilComputationSDFGBuilder.SDFGContext",
        node_ctx: "StencilComputationSDFGBuilder.NodeContext",
        symtable: ChainMap[eve.SymbolRef, dcir.Decl],
        **kwargs,
    ) -> None:
        code = TaskletCodegen.apply_codegen(
            node,
            read_memlets=node.read_memlets,
            write_memlets=node.write_memlets,
            sdfg_ctx=sdfg_ctx,
            symtable=symtable,
        )

        tasklet = sdfg_ctx.state.add_tasklet(
            name=f"{sdfg_ctx.sdfg.label}_Tasklet",
            code=code,
            inputs=set(memlet.connector for memlet in node.read_memlets),
            outputs=set(memlet.connector for memlet in node.write_memlets),
            debuginfo=get_dace_debuginfo(node),
        )

        self.visit(
            node.read_memlets,
            scope_node=tasklet,
            sdfg_ctx=sdfg_ctx,
            node_ctx=node_ctx,
            symtable=symtable,
            **kwargs,
        )
        self.visit(
            node.write_memlets,
            scope_node=tasklet,
            sdfg_ctx=sdfg_ctx,
            node_ctx=node_ctx,
            symtable=symtable,
            **kwargs,
        )
        StencilComputationSDFGBuilder._add_empty_edges(
            tasklet, tasklet, sdfg_ctx=sdfg_ctx, node_ctx=node_ctx
        )

    def visit_Range(self, node: dcir.Range, **kwargs) -> Dict[str, str]:
        start, end = node.interval.to_dace_symbolic()
        return {node.var: str(dace.subsets.Range([(start, end - 1, node.stride)]))}

    def visit_DomainMap(
        self,
        node: dcir.DomainMap,
        *,
        node_ctx: "StencilComputationSDFGBuilder.NodeContext",
        sdfg_ctx: "StencilComputationSDFGBuilder.SDFGContext",
        **kwargs,
    ) -> None:
        ndranges = {
            k: v
            for index_range in node.index_ranges
            for k, v in self.visit(index_range, **kwargs).items()
        }
        name = sdfg_ctx.sdfg.label + "".join(ndranges.keys()) + "_map"
        map_entry, map_exit = sdfg_ctx.state.add_map(
            name=name,
            ndrange=ndranges,
            schedule=node.schedule.to_dace_schedule(),
            debuginfo=get_dace_debuginfo(node),
        )

        for scope_node in node.computations:
            input_node_and_conns: Dict[Optional[str], Tuple[dace.nodes.Node, Optional[str]]] = {}
            output_node_and_conns: Dict[Optional[str], Tuple[dace.nodes.Node, Optional[str]]] = {}
            for field in set(memlet.field for memlet in scope_node.read_memlets):
                map_entry.add_in_connector("IN_" + field)
                map_entry.add_out_connector("OUT_" + field)
                input_node_and_conns[field] = (map_entry, "OUT_" + field)
            for field in set(memlet.field for memlet in scope_node.write_memlets):
                map_exit.add_in_connector("IN_" + field)
                map_exit.add_out_connector("OUT_" + field)
                output_node_and_conns[field] = (map_exit, "IN_" + field)
            if not input_node_and_conns:
                input_node_and_conns[None] = (map_entry, None)
            if not output_node_and_conns:
                output_node_and_conns[None] = (map_exit, None)
            inner_node_ctx = StencilComputationSDFGBuilder.NodeContext(
                input_node_and_conns=input_node_and_conns,
                output_node_and_conns=output_node_and_conns,
            )
            self.visit(scope_node, sdfg_ctx=sdfg_ctx, node_ctx=inner_node_ctx, **kwargs)

        self.visit(
            node.read_memlets,
            scope_node=map_entry,
            sdfg_ctx=sdfg_ctx,
            node_ctx=node_ctx,
            connector_prefix="IN_",
            **kwargs,
        )
        self.visit(
            node.write_memlets,
            scope_node=map_exit,
            sdfg_ctx=sdfg_ctx,
            node_ctx=node_ctx,
            connector_prefix="OUT_",
            **kwargs,
        )
        StencilComputationSDFGBuilder._add_empty_edges(
            map_entry, map_exit, sdfg_ctx=sdfg_ctx, node_ctx=node_ctx
        )

    def visit_DomainLoop(
        self,
        node: dcir.DomainLoop,
        *,
        sdfg_ctx: "StencilComputationSDFGBuilder.SDFGContext",
        **kwargs,
    ) -> None:
        sdfg_ctx = sdfg_ctx.add_loop(node.index_range)
        self.visit(node.loop_states, sdfg_ctx=sdfg_ctx, **kwargs)
        sdfg_ctx.pop_loop()

    def visit_ComputationState(
        self,
        node: dcir.ComputationState,
        *,
        sdfg_ctx: "StencilComputationSDFGBuilder.SDFGContext",
        **kwargs,
    ) -> None:

        sdfg_ctx.add_state()
        read_acc_and_conn: Dict[Optional[str], Tuple[dace.nodes.Node, Optional[str]]] = {}
        write_acc_and_conn: Dict[Optional[str], Tuple[dace.nodes.Node, Optional[str]]] = {}
        for computation in node.computations:
            assert isinstance(computation, dcir.ComputationNode)
            for memlet in computation.read_memlets:
                if memlet.field not in read_acc_and_conn:
                    read_acc_and_conn[memlet.field] = (
                        sdfg_ctx.state.add_access(memlet.field, debuginfo=dace.DebugInfo(0)),
                        None,
                    )
            for memlet in computation.write_memlets:
                if memlet.field not in write_acc_and_conn:
                    write_acc_and_conn[memlet.field] = (
                        sdfg_ctx.state.add_access(memlet.field, debuginfo=dace.DebugInfo(0)),
                        None,
                    )
            node_ctx = StencilComputationSDFGBuilder.NodeContext(
                input_node_and_conns=read_acc_and_conn, output_node_and_conns=write_acc_and_conn
            )
            self.visit(computation, sdfg_ctx=sdfg_ctx, node_ctx=node_ctx, **kwargs)

    def visit_FieldDecl(
        self,
        node: dcir.FieldDecl,
        *,
        sdfg_ctx: "StencilComputationSDFGBuilder.SDFGContext",
        non_transients: Set[eve.SymbolRef],
        **kwargs,
    ) -> None:
        assert len(node.strides) == len(node.shape)
        sdfg_ctx.sdfg.add_array(
            node.name,
            shape=node.shape,
            strides=[dace.symbolic.pystr_to_symbolic(s) for s in node.strides],
            dtype=data_type_to_dace_typeclass(node.dtype),
            storage=node.storage.to_dace_storage(),
            transient=node.name not in non_transients,
            debuginfo=dace.DebugInfo(0),
        )

    def visit_SymbolDecl(
        self,
        node: dcir.SymbolDecl,
        *,
        sdfg_ctx: "StencilComputationSDFGBuilder.SDFGContext",
        **kwargs,
    ) -> None:
        if node.name not in sdfg_ctx.sdfg.symbols:
            sdfg_ctx.sdfg.add_symbol(node.name, stype=data_type_to_dace_typeclass(node.dtype))

    def visit_NestedSDFG(
        self,
        node: dcir.NestedSDFG,
        *,
        sdfg_ctx: Optional["StencilComputationSDFGBuilder.SDFGContext"] = None,
        node_ctx: Optional["StencilComputationSDFGBuilder.NodeContext"] = None,
        symtable: ChainMap[eve.SymbolRef, Any],
        **kwargs,
    ) -> dace.nodes.NestedSDFG:

        sdfg = dace.SDFG(node.label)
        inner_sdfg_ctx = StencilComputationSDFGBuilder.SDFGContext(
            sdfg=sdfg, state=sdfg.add_state(is_start_state=True)
        )
        self.visit(
            node.field_decls,
            sdfg_ctx=inner_sdfg_ctx,
            non_transients={memlet.connector for memlet in node.read_memlets + node.write_memlets},
            **kwargs,
        )
        self.visit(node.symbol_decls, sdfg_ctx=inner_sdfg_ctx, **kwargs)
        symbol_mapping = {decl.name: decl.to_dace_symbol() for decl in node.symbol_decls}

        for computation_state in node.states:
            self.visit(computation_state, sdfg_ctx=inner_sdfg_ctx, symtable=symtable, **kwargs)

        if sdfg_ctx is not None and node_ctx is not None:
            nsdfg = sdfg_ctx.state.add_nested_sdfg(
                sdfg=sdfg,
                parent=None,
                inputs=node.input_connectors,
                outputs=node.output_connectors,
                symbol_mapping=symbol_mapping,
                debuginfo=dace.DebugInfo(0),
            )
            self.visit(
                node.read_memlets,
                scope_node=nsdfg,
                sdfg_ctx=sdfg_ctx,
                node_ctx=node_ctx,
                symtable=symtable.parents,
                **kwargs,
            )
            self.visit(
                node.write_memlets,
                scope_node=nsdfg,
                sdfg_ctx=sdfg_ctx,
                node_ctx=node_ctx,
                symtable=symtable.parents,
                **kwargs,
            )
            StencilComputationSDFGBuilder._add_empty_edges(
                nsdfg, nsdfg, sdfg_ctx=sdfg_ctx, node_ctx=node_ctx
            )
        else:
            nsdfg = dace.nodes.NestedSDFG(
                label=sdfg.label,
                sdfg=sdfg,
                inputs={memlet.connector for memlet in node.read_memlets},
                outputs={memlet.connector for memlet in node.write_memlets},
                symbol_mapping=symbol_mapping,
            )

        return nsdfg



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import typing

import attr


class _TypeDescriptor:
    def __init__(self, name, args, make_validator_func, type_hint):
        self.name = name
        assert args is None or isinstance(args, tuple)
        self.args = args
        self.make_validator_func = make_validator_func
        self.type_hint = type_hint

    def __repr__(self):
        args = self.args if self.args is not None else []
        arg_names = []
        for a in args:
            if isinstance(a, type):
                arg_names.append(a.__name__)
            elif isinstance(a, _TypeDescriptor):
                arg_names.append(repr(a))
        args = "[{}]".format(", ".join(arg_names)) if len(arg_names) > 0 else ""
        return "{}{}".format(self.name, args)

    @property
    def validator(self):
        return self.make_validator_func(self.args)


class _GenericTypeDescriptor:
    def __init__(self, name, n_args, make_validator_func, generic_type_hint):
        self.name = name

        if not isinstance(n_args, tuple):
            n_args = (n_args, n_args)
        assert all(isinstance(n, int) and n > 0 for n in n_args if n is not None)

        self.n_args = n_args

        assert callable(make_validator_func)
        self.make_validator_func = make_validator_func
        self.generic_type_hint = generic_type_hint

    def __getitem__(self, type_list):
        if not isinstance(type_list, tuple):
            type_list = tuple([type_list])
        assert all(isinstance(t, (type, _TypeDescriptor)) for t in type_list)
        assert (self.n_args[0] is None or self.n_args[0] <= len(type_list)) and (
            self.n_args[1] is None or self.n_args[1] <= self.n_args[1]
        )
        hint_list = []
        for t in type_list:
            if isinstance(t, type):
                hint_list.append(t)
            else:
                hint_list.append(getattr(t, "type_hint", typing.Any))

        return _TypeDescriptor(
            self.name,
            args=tuple(type_list),
            make_validator_func=self.make_validator_func,
            type_hint=self.generic_type_hint[tuple(hint_list)],
        )


def _make_type_validator(t):
    assert isinstance(t, (type, _TypeDescriptor))

    if isinstance(t, type):
        return attr.validators.instance_of(t)
    else:
        # isinstance(t, _TypeDescriptor)
        return t.validator


def _make_sequence_validator(type_list, container_types=(list, tuple)):
    assert type_list is not None and len(type_list) == 1
    item_type = type_list[0]

    item_validator = _make_type_validator(item_type)

    def _is_sequence_of_validator(instance, attribute, value):
        try:
            assert isinstance(value, tuple(container_types))
            assert isinstance([item_validator(instance, attribute, v) for v in value], list)
        except Exception:
            raise ValueError(
                "Expr ({value}) does not match the '{name}' specification".format(
                    value=value, name=attribute.name
                )
            )

    return _is_sequence_of_validator


def _make_list_validator(type_list):
    return _make_sequence_validator(type_list, container_types=[list])


def _make_set_validator(type_list):
    return _make_sequence_validator(type_list, container_types=[set])


def _make_dict_validator(type_list):
    assert type_list is not None and len(type_list) == 2
    key_validator, value_validator = [_make_type_validator(t) for t in type_list]

    def _is_dict_of_validator(instance, attribute, value):
        try:
            assert isinstance(value, dict)
            assert isinstance([key_validator(instance, attribute, v) for v in value.keys()], list)
            assert isinstance(
                [value_validator(instance, attribute, v) for v in value.values()], list
            )
        except Exception:
            raise ValueError(
                "Expr ({value}) does not match the '{name}' specification".format(
                    value=value, name=attribute.name
                )
            )

    return _is_dict_of_validator


def _make_tuple_validator(type_list):
    assert type_list is not None and len(type_list) > 0
    validators = [_make_type_validator(t) for t in type_list]

    def _is_tuple_of_validator(instance, attribute, value):
        try:
            assert isinstance(value, tuple)
            assert len(value) == len(validators)
            assert isinstance(
                [
                    validator(instance, attribute, value)
                    for value, validator in zip(value, validators)
                ],
                list,
            )
        except Exception:
            raise ValueError(
                "Expr ({value}) does not match the '{name}' specification".format(
                    value=value, name=attribute.name
                )
            )

    return _is_tuple_of_validator


def _make_union_validator(type_list):
    assert type_list is not None and len(type_list) > 1
    validators = [_make_type_validator(t) for t in type_list]

    def _is_union_of_validator(instance, attribute, value):
        passed = True
        for validator in validators:
            try:
                validator(instance, attribute, value)
                break

            except Exception as e:
                pass
        else:
            passed = False

        if not passed:
            raise ValueError(
                "Expr ({value}) does not match the '{name}' specification".format(
                    value=value, name=attribute.name
                )
            )

    return _is_union_of_validator


def _make_any_validator(type_list):
    def _is_any_validator(instance, attribute, value):
        pass

    return _is_any_validator


def _make_nothing_validator(type_list):
    def _is_nothing_validator(instance, attribute, value):
        assert value is None

    return _is_nothing_validator


Any = _TypeDescriptor("Any", None, _make_any_validator, typing.Any)

Sequence = _GenericTypeDescriptor("Sequence", 1, _make_sequence_validator, typing.Sequence)
List = _GenericTypeDescriptor("List", 1, _make_list_validator, typing.List)
Dict = _GenericTypeDescriptor("Dict", 2, _make_dict_validator, typing.Dict)
Set = _GenericTypeDescriptor("Set", 1, _make_set_validator, typing.Set)
Tuple = _GenericTypeDescriptor("Tuple", (1, None), _make_tuple_validator, typing.Tuple)
Union = _GenericTypeDescriptor("Union", (2, None), _make_union_validator, typing.Union)

Optional = _GenericTypeDescriptor(
    "Optional",
    (1, None),
    lambda type_list: _make_union_validator([*type_list, type(None)]),
    typing.Optional,
)


def attribute(of, optional=False, **kwargs):
    if isinstance(of, _TypeDescriptor):
        attr_validator = of.validator
        attr_type_hint = of.type_hint

    elif isinstance(of, type):
        # assert of in (bool, float, str, int, enum.Enum)
        attr_validator = attr.validators.instance_of(of)
        attr_type_hint = of

    else:
        raise ValueError("Invalid attribute type '{}'".format(of))

    if optional:
        attr_validator = attr.validators.optional(attr_validator)
        kwargs.setdefault("default", None)
        # kwargs["kw_only"] = True

    return attr.ib(validator=attr_validator, type=attr_type_hint, **kwargs)


class AttributeClassLike:
    def validate(self):
        ...

    @property
    def attributes(self):
        ...

    @property
    def as_dict(self):
        ...


def attribclass(cls_or_none=None, **kwargs):
    """Class decorator to convert a regular class into an `AttribClass`."""

    def validate(self):
        """Validate this instance's data attributes."""
        attr.validate(self)

    def attributes(self):
        """Generate a :class:`list` with this class' attribute names."""
        result = [a.name for a in attr.fields(self.__class__)]
        return result

    def as_dict(self):
        """Generate a :class:`dict` with this instance's data attributes."""
        return attr.asdict(self)

    extra_members = dict(validate=validate, attributes=property(attributes), as_dict=as_dict)

    def _make_attrs_class_wrapper(cls):
        for name, member in extra_members.items():
            if name in cls.__dict__.keys():
                raise ValueError(
                    "Name clashing with a existing '{name}' member"
                    " of the decorated class ".format(name=name)
                )
            setattr(cls, name, member)

        new_cls = attr.s(cls, **kwargs)
        return new_cls

    if cls_or_none is None:
        return _make_attrs_class_wrapper
    else:
        return _make_attrs_class_wrapper(cls_or_none)


def attribkwclass(cls_or_none=None, **kwargs):
    kwargs.setdefault("kw_only", True)
    return attribclass(cls_or_none=cls_or_none, **kwargs)


def attributes_of(an_attribclass):
    return [a.name for a in attr.fields(an_attribclass)]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
GTScript import machinery.

Usage Example
-------------

.. code-block::

    from gt4py.cartesian import gtscript_imports

    # simple usage
    gtscript_imports.enable()  # for allowing .gt.py everywhere in sys.path

    # advanced usage
    gtscript_imports.enable(
        search_path=[<path1>, <path2>, ...],  # for allowing only in search_path
        generate_path=<mybuildpath>,  # for generating python modules in a specific dir
        in_source=False,  # set True to generate python modules next to gtscfipt files
    )

    # scoped usage
    with gtscript_imports.enabled():
        import ...

"""
import importlib
import importlib.abc
import pathlib
import sys
import tempfile
from contextlib import contextmanager
from types import ModuleType
from typing import Any, Generator, Iterator, List, Optional, Union


GTS_EXTENSIONS = [".gt.py"]
GTS_COMMENT = "# [GT] using-dsl: gtscript"
GTS_IMPORT = (
    "from gt4py.cartesian.gtscript import *  "
    "# This file is automatically generated, changes may be overwritten."
)


def _add_extension(path: pathlib.Path, extension: str) -> pathlib.Path:
    """Add an extension to the filename of a path."""
    return path.parent / (path.name + extension)


class GtsFinder(importlib.abc.MetaPathFinder):
    """
    Implements the :py:class:`importlib.abc.PathFinder` protocol.

    This finder is responsible for finding GTScript files within a set of
    search paths.

    Parameters
    ----------
    search_path :
        Search path for `gtscript` sources, defaults to `sys.path`

    generate_path :
        Path to generate py modules in. use a temp dir by default

    in_source :
        If True, py modules are built next to the gtscript files
        (generate_path is ignored).
    """

    def __init__(
        self,
        search_path: Optional[List[Union[str, pathlib.Path]]] = None,
        generate_path: Optional[Union[str, pathlib.Path]] = None,
        in_source: bool = False,
    ):
        if in_source:
            self.generate_path = None
        elif generate_path:
            self.generate_path = pathlib.Path(generate_path)
        else:
            self.generate_path = pathlib.Path(tempfile.gettempdir())

        self.search_path = search_path

    def get_generate_path(self, src_file_path: pathlib.Path) -> pathlib.Path:
        """Find the out-of-source or in-source generate directory."""
        return self.generate_path or src_file_path.parent

    def iter_search_candidates(
        self, fullname: str, path: Optional[List[pathlib.Path]]
    ) -> Generator[pathlib.Path, None, None]:
        """Iterate possible source file paths."""
        search_paths = [p for p in self.search_path or sys.path]
        search_paths.extend(path or [])
        filepath = pathlib.Path(fullname.rsplit(".")[-1])
        for search_path in search_paths:
            search_path = pathlib.Path(search_path)
            for extension in GTS_EXTENSIONS:
                yield search_path.absolute() / _add_extension(filepath, extension)

    def find_spec(
        self, fullname: str, path=None, target=None
    ) -> Optional[importlib.machinery.ModuleSpec]:
        """Create a module spec for the first matching source file path."""
        if fullname in sys.modules:
            return None

        for candidate in self.iter_search_candidates(fullname, path):
            if candidate.exists():
                spec = importlib.machinery.ModuleSpec(
                    name=fullname,
                    loader=GtsLoader(
                        fullname, candidate, generate_at=self.get_generate_path(candidate)
                    ),
                    origin=str(candidate),
                    is_package=False,
                )
                spec.has_location = True
                return spec

        return None

    def install(self) -> None:
        sys.meta_path.append(self)
        if self.search_path:
            sys.path.extend([str(p) for p in self.search_path])

    def load_module(self, fullname: str, path=None, target=None) -> ModuleType:
        if fullname in sys.modules:
            return sys.modules[fullname]
        spec = self.find_spec(fullname, path, target)
        if not spec or not isinstance(spec.loader, GtsLoader):
            raise ImportError("could not find gtscript module %s", fullname)
        module = spec.loader.create_module(spec)
        spec.loader.exec_module(module)
        return module


class GtsLoader(importlib.machinery.SourceFileLoader):
    """
    Extend :py:class:`importlib.machinery.SourceFileLoader` for GTScript files.

    Generate a python module for a GTScript file and use the super class to
    load that instead.
    """

    def __init__(self, fullname: str, path: pathlib.Path, generate_at: pathlib.Path):
        self.module_file = generate_at / (path.stem.split(".")[0] + ".py")
        super().__init__(fullname, str(path.absolute()))

    @property
    def plpath(self) -> pathlib.Path:
        return pathlib.Path(str(self.path))

    def get_filename(self, fullname: Optional[str] = None) -> str:
        """
        Generate a py module if an up to date one doesn't exist yet.

        Create the generation directory if necessary.
        Use file stats to check if the module needs to be updated.

        Parameters
        ----------
        fullname : `str`, optional
            Dotted name corresponding to the gtscript module.

        Returns
        -------
            The file path of the generated py module as a string
        """
        if not self.module_file.parent.exists():
            self.module_file.parent.mkdir(exist_ok=True)
        if not self.module_file.exists():
            self.module_file.touch()

        if self.path_stats(self.path) != self.path_stats(str(self.module_file.absolute())):
            assert fullname is not None
            self.module_file.write_text(self.get_source_code(fullname))

        return str(self.module_file)

    def get_source_code(self, fullname: str) -> str:
        return self.plpath.read_text().replace(GTS_COMMENT, GTS_IMPORT)

    def create_module(self, spec: importlib.machinery.ModuleSpec) -> ModuleType:
        module = ModuleType(name=spec.name)
        return module


def enable(
    *,
    search_path: Optional[List[Union[str, pathlib.Path]]] = None,
    generate_path: Optional[Union[str, pathlib.Path]] = None,
    in_source: bool = False,
) -> GtsFinder:
    """
    Install GTScript import extensions.

    Parameters are passed through to the constructor of :py:class:`GtsFinder`.
    """
    finder = GtsFinder(search_path=search_path, generate_path=generate_path, in_source=in_source)
    finder.install()
    return finder


@contextmanager
def enabled(**kwargs: Any) -> Iterator:
    """
    Create a context within which GTScript extensions can be imported.

    Example
    -------
    .. code-block: python

        mystencil = None
        with gtscript_imports.enabled(search_path=["my/gtscript/extensions/"]):
            import some_stencil  # works
            mystencil = some_stencil.some_stencil

        ## use mystencil

        import some_other_stencil  # in the same directory as some_stencil.gt.py raises error
    """
    backup_import_system = (
        sys.path.copy(),
        sys.meta_path.copy(),
        sys.modules.copy(),
    )
    try:
        yield enable(**kwargs)
    finally:
        sys.path, sys.meta_path, sys.modules = backup_import_system



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import functools
import itertools
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Set, Union, cast

from devtools import debug  # noqa: F401
from typing_extensions import Protocol

from gt4py import eve
from gt4py.cartesian.gtc import common, oir
from gt4py.cartesian.gtc.common import CartesianOffset, ExprKind
from gt4py.cartesian.gtc.gtcpp import gtcpp
from gt4py.cartesian.gtc.passes.oir_optimizations.utils import (
    collect_symbol_names,
    symbol_name_creator,
)


# - Each HorizontalExecution is a Functor (and a Stage)
# - Each VerticalLoop is MultiStage


def _extract_accessors(node: eve.Node, temp_names: Set[str]) -> List[gtcpp.GTAccessor]:
    extents = (
        node.walk_values()
        .if_isinstance(gtcpp.AccessorRef)
        .reduceby(
            (lambda extent, accessor_ref: extent + accessor_ref.offset),
            "name",
            init=gtcpp.GTExtent.zero(),
            as_dict=True,
        )
    )

    inout_fields: Set[str] = (
        node.walk_values()
        .if_isinstance(gtcpp.AssignStmt)
        .getattr("left")
        .if_isinstance(gtcpp.AccessorRef)
        .getattr("name")
        .to_set()
    )
    ndims = dict(
        node.walk_values()
        .if_isinstance(gtcpp.AccessorRef)
        .map(
            lambda accessor: (
                accessor.name,
                3 + (len(accessor.data_index) if accessor.name not in temp_names else 0),
            )
        )
    )

    return [
        gtcpp.GTAccessor(
            name=name,
            id=i,
            intent=gtcpp.Intent.INOUT if name in inout_fields else gtcpp.Intent.IN,
            extent=extent,
            ndim=ndims[name],
        )
        for i, (name, extent) in enumerate(extents.items())
    ]


def _make_axis_offset_expr(
    bound: common.AxisBound,
    axis_index: int,
    axis_length_accessor: Callable[[int], gtcpp.AccessorRef],
) -> gtcpp.Expr:
    if bound.level == common.LevelMarker.END:
        base = axis_length_accessor(axis_index)
        return gtcpp.BinaryOp(
            op=common.ArithmeticOperator.ADD,
            left=base,
            right=gtcpp.Literal(value=str(bound.offset), dtype=common.DataType.INT32),
        )
    else:
        return gtcpp.Literal(value=str(bound.offset), dtype=common.DataType.INT32)


class SymbolNameCreator(Protocol):
    def __call__(self, name: str) -> str:
        ...


class OIRToGTCpp(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    @dataclass
    class ProgramContext:
        functors: List[gtcpp.GTFunctor] = field(default_factory=list)

        def add_functor(self, functor: gtcpp.GTFunctor) -> "OIRToGTCpp.ProgramContext":
            self.functors.append(functor)
            return self

    @dataclass
    class GTComputationContext:
        create_symbol_name: SymbolNameCreator
        temporaries: List[gtcpp.Temporary] = field(default_factory=list)
        positionals: Dict[int, gtcpp.Positional] = field(default_factory=dict)
        axis_lengths: Dict[int, gtcpp.AxisLength] = field(default_factory=dict)
        _arguments: Set[str] = field(default_factory=set)

        def add_temporaries(
            self, temporaries: List[gtcpp.Temporary]
        ) -> "OIRToGTCpp.GTComputationContext":
            self.temporaries.extend(temporaries)
            return self

        @property
        def arguments(self) -> List[gtcpp.Arg]:
            return [gtcpp.Arg(name=name) for name in self._arguments]

        def add_arguments(self, arguments: Set[str]) -> "OIRToGTCpp.GTComputationContext":
            self._arguments.update(arguments)
            return self

        @staticmethod
        def _make_scalar_accessor(name: str) -> gtcpp.AccessorRef:
            return gtcpp.AccessorRef(
                name=name,
                offset=CartesianOffset.zero(),
                kind=ExprKind.SCALAR,
                dtype=common.DataType.INT32,
            )

        def make_positional(self, axis: int) -> gtcpp.AccessorRef:
            axis_name = ["I", "J", "K"][axis].lower()
            name = self.create_symbol_name(f"ax{axis}_ind")
            positional = self.positionals.setdefault(
                axis, gtcpp.Positional(name=name, axis_name=axis_name)
            )
            return self._make_scalar_accessor(positional.name)

        def make_length(self, axis: int) -> gtcpp.AccessorRef:
            name = self.create_symbol_name(f"ax{axis}_len")
            length = self.axis_lengths.setdefault(axis, gtcpp.AxisLength(name=name, axis=axis))
            return self._make_scalar_accessor(length.name)

        @property
        def extra_decls(self) -> List[gtcpp.ComputationDecl]:
            return list(self.positionals.values()) + list(self.axis_lengths.values())

    def visit_Literal(self, node: oir.Literal, **kwargs: Any) -> gtcpp.Literal:
        return gtcpp.Literal(value=node.value, dtype=node.dtype)

    def visit_UnaryOp(self, node: oir.UnaryOp, **kwargs: Any) -> gtcpp.UnaryOp:
        return gtcpp.UnaryOp(op=node.op, expr=self.visit(node.expr, **kwargs))

    def visit_BinaryOp(self, node: oir.BinaryOp, **kwargs: Any) -> gtcpp.BinaryOp:
        return gtcpp.BinaryOp(
            op=node.op,
            left=self.visit(node.left, **kwargs),
            right=self.visit(node.right, **kwargs),
        )

    def visit_TernaryOp(self, node: oir.TernaryOp, **kwargs: Any) -> gtcpp.TernaryOp:
        return gtcpp.TernaryOp(
            cond=self.visit(node.cond, **kwargs),
            true_expr=self.visit(node.true_expr, **kwargs),
            false_expr=self.visit(node.false_expr, **kwargs),
        )

    def visit_NativeFuncCall(self, node: oir.NativeFuncCall, **kwargs: Any) -> gtcpp.NativeFuncCall:
        return gtcpp.NativeFuncCall(func=node.func, args=self.visit(node.args, **kwargs))

    def visit_Cast(self, node: oir.Cast, **kwargs: Any) -> gtcpp.Cast:
        return gtcpp.Cast(dtype=node.dtype, expr=self.visit(node.expr, **kwargs))

    def visit_Temporary(self, node: oir.Temporary, **kwargs: Any) -> gtcpp.Temporary:
        return gtcpp.Temporary(name=node.name, dtype=node.dtype, data_dims=node.data_dims)

    def visit_VariableKOffset(
        self, node: oir.VariableKOffset, **kwargs: Any
    ) -> gtcpp.VariableKOffset:
        return gtcpp.VariableKOffset(k=self.visit(node.k, **kwargs))

    def visit_FieldAccess(self, node: oir.FieldAccess, **kwargs: Any) -> gtcpp.AccessorRef:
        return gtcpp.AccessorRef(
            name=node.name,
            offset=self.visit(node.offset, **kwargs),
            data_index=self.visit(node.data_index, **kwargs),
            dtype=node.dtype,
        )

    def visit_ScalarAccess(
        self, node: oir.ScalarAccess, **kwargs: Any
    ) -> Union[gtcpp.AccessorRef, gtcpp.LocalAccess]:
        assert "symtable" in kwargs
        if node.name in kwargs["symtable"]:
            symbol = kwargs["symtable"][node.name]
            if isinstance(symbol, oir.ScalarDecl):
                return gtcpp.AccessorRef(
                    name=symbol.name, offset=CartesianOffset.zero(), dtype=symbol.dtype
                )
            assert isinstance(symbol, oir.LocalScalar)
        return gtcpp.LocalAccess(name=node.name, dtype=node.dtype)

    def visit_AxisBound(
        self, node: oir.AxisBound, *, is_start: bool, **kwargs: Any
    ) -> gtcpp.GTLevel:
        if node.level == common.LevelMarker.START:
            splitter = 0
            offset = node.offset + 1 if (node.offset >= 0 and is_start) else node.offset
        elif node.level == common.LevelMarker.END:
            splitter = 1
            offset = node.offset - 1 if (node.offset <= 0 and not is_start) else node.offset
        else:
            raise ValueError("Cannot handle dynamic levels")
        return gtcpp.GTLevel(splitter=splitter, offset=offset)

    def visit_Interval(self, node: oir.Interval, **kwargs: Any) -> gtcpp.GTInterval:
        return gtcpp.GTInterval(
            from_level=self.visit(node.start, is_start=True),
            to_level=self.visit(node.end, is_start=False),
        )

    def _mask_to_expr(
        self, mask: common.HorizontalMask, comp_ctx: "GTComputationContext"
    ) -> gtcpp.Expr:
        mask_expr: List[gtcpp.Expr] = []
        for axis_index, interval in enumerate(mask.intervals):
            if interval.is_single_index():
                assert interval.start is not None
                mask_expr.append(
                    gtcpp.BinaryOp(
                        op=common.ComparisonOperator.EQ,
                        left=comp_ctx.make_positional(axis_index),
                        right=_make_axis_offset_expr(
                            interval.start, axis_index, comp_ctx.make_length
                        ),
                    )
                )
            else:
                for op, endpt in zip(
                    (common.ComparisonOperator.GE, common.ComparisonOperator.LT),
                    (interval.start, interval.end),
                ):
                    if endpt is None:
                        continue
                    mask_expr.append(
                        gtcpp.BinaryOp(
                            op=op,
                            left=comp_ctx.make_positional(axis_index),
                            right=_make_axis_offset_expr(endpt, axis_index, comp_ctx.make_length),
                        )
                    )
        return (
            functools.reduce(
                lambda a, b: gtcpp.BinaryOp(op=common.LogicalOperator.AND, left=a, right=b),
                mask_expr,
            )
            if mask_expr
            else gtcpp.Literal(value=common.BuiltInLiteral.TRUE, dtype=common.DataType.BOOL)
        )

    def visit_HorizontalRestriction(
        self, node: oir.HorizontalRestriction, **kwargs: Any
    ) -> gtcpp.IfStmt:
        mask = self._mask_to_expr(node.mask, kwargs["comp_ctx"])
        return gtcpp.IfStmt(
            cond=mask, true_branch=gtcpp.BlockStmt(body=self.visit(node.body, **kwargs))
        )

    def visit_AssignStmt(self, node: oir.AssignStmt, **kwargs: Any) -> gtcpp.AssignStmt:
        assert "symtable" in kwargs
        return gtcpp.AssignStmt(
            left=self.visit(node.left, **kwargs), right=self.visit(node.right, **kwargs)
        )

    def visit_MaskStmt(self, node: oir.MaskStmt, **kwargs: Any) -> gtcpp.IfStmt:
        return gtcpp.IfStmt(
            cond=self.visit(node.mask, **kwargs),
            true_branch=gtcpp.BlockStmt(body=self.visit(node.body, **kwargs)),
        )

    def visit_While(self, node: oir.While, **kwargs: Any) -> gtcpp.While:
        return gtcpp.While(
            cond=self.visit(node.cond, **kwargs), body=self.visit(node.body, **kwargs)
        )

    def visit_HorizontalExecution(
        self,
        node: oir.HorizontalExecution,
        *,
        prog_ctx: "ProgramContext",
        comp_ctx: "GTComputationContext",
        interval: gtcpp.GTInterval,
        **kwargs: Any,
    ) -> gtcpp.GTStage:
        assert "symtable" in kwargs
        apply_method = gtcpp.GTApplyMethod(
            interval=self.visit(interval, **kwargs),
            body=self.visit(node.body, comp_ctx=comp_ctx, **kwargs),
            local_variables=self.visit(node.declarations, **kwargs),
        )
        accessors = _extract_accessors(apply_method, {decl.name for decl in comp_ctx.temporaries})
        stage_args = [gtcpp.Arg(name=acc.name) for acc in accessors]

        tmp_names = {tmp.name for tmp in comp_ctx.temporaries}
        param_names_not_tmps = {
            str(param_arg.name) for param_arg in stage_args if param_arg.name not in tmp_names
        }

        comp_ctx.add_arguments(param_names_not_tmps)

        functor_name = type(node).__name__ + str(id(node))
        prog_ctx.add_functor(
            gtcpp.GTFunctor(
                name=functor_name,
                applies=[apply_method],
                param_list=gtcpp.GTParamList(accessors=accessors),
            )
        )

        return gtcpp.GTStage(functor=functor_name, args=stage_args)

    def visit_VerticalLoop(
        self,
        node: oir.VerticalLoop,
        *,
        comp_ctx: GTComputationContext,
        **kwargs: Any,
    ) -> gtcpp.GTMultiStage:
        # the following visit assumes that temporaries are already available in comp_ctx
        stages = list(
            itertools.chain(
                *(
                    self.visit(
                        section.horizontal_executions,
                        interval=section.interval,
                        default=([], []),
                        comp_ctx=comp_ctx,
                        **kwargs,
                    )
                    for section in node.sections
                )
            )
        )
        caches = self.visit(node.caches)
        return gtcpp.GTMultiStage(loop_order=node.loop_order, stages=stages, caches=caches)

    def visit_IJCache(self, node: oir.IJCache, **kwargs: Any) -> gtcpp.IJCache:
        return gtcpp.IJCache(name=node.name, loc=node.loc)

    def visit_KCache(self, node: oir.KCache, **kwargs: Any) -> gtcpp.KCache:
        return gtcpp.KCache(name=node.name, fill=node.fill, flush=node.flush, loc=node.loc)

    def visit_FieldDecl(self, node: oir.FieldDecl, **kwargs: Any) -> gtcpp.FieldDecl:
        return gtcpp.FieldDecl(
            name=node.name, dtype=node.dtype, dimensions=node.dimensions, data_dims=node.data_dims
        )

    def visit_ScalarDecl(self, node: oir.ScalarDecl, **kwargs: Any) -> gtcpp.GlobalParamDecl:
        return gtcpp.GlobalParamDecl(name=node.name, dtype=node.dtype)

    def visit_LocalScalar(self, node: oir.LocalScalar, **kwargs: Any) -> gtcpp.LocalVarDecl:
        return gtcpp.LocalVarDecl(name=node.name, dtype=node.dtype, loc=node.loc)

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> gtcpp.Program:
        prog_ctx = self.ProgramContext()
        comp_ctx = self.GTComputationContext(
            create_symbol_name=cast(
                SymbolNameCreator, symbol_name_creator(collect_symbol_names(node))
            )
        )

        assert all([isinstance(decl, oir.Temporary) for decl in node.declarations])
        comp_ctx.add_temporaries(self.visit(node.declarations))

        multi_stages = self.visit(
            node.vertical_loops, prog_ctx=prog_ctx, comp_ctx=comp_ctx, **kwargs
        )

        gt_computation = gtcpp.GTComputationCall(
            arguments=comp_ctx.arguments,
            extra_decls=comp_ctx.extra_decls,
            temporaries=comp_ctx.temporaries,
            multi_stages=multi_stages,
        )
        parameters = self.visit(node.params)
        return gtcpp.Program(
            name=node.name,
            parameters=parameters,
            functors=prog_ctx.functors,
            gt_computation=gt_computation,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import contextlib
import copy
import distutils
import distutils.sysconfig
import io
import os
import shutil
from typing import Any, Dict, List, Optional, Tuple, Type, Union, overload

import pybind11
import setuptools
from setuptools.command.build_ext import build_ext

from gt4py.cartesian import config as gt_config


def get_dace_module_path() -> Optional[str]:
    try:
        import dace

        return os.path.dirname(dace.__file__)
    except ImportError:
        return None


def get_cuda_compute_capability():
    try:
        import cupy as cp

        return cp.cuda.Device(0).compute_capability
    except ImportError:
        return None


def get_gt_pyext_build_opts(
    *,
    debug_mode: bool = False,
    add_profile_info: bool = False,
    uses_openmp: bool = True,
    uses_cuda: bool = False,
    gt_version: int = 1,
) -> Dict[str, Union[str, List[str], Dict[str, Any]]]:

    include_dirs = [gt_config.build_settings["boost_include_path"]]
    extra_compile_args_from_config = gt_config.build_settings["extra_compile_args"]

    if uses_cuda:
        compute_capability = get_cuda_compute_capability()
        cuda_arch = gt_config.build_settings["cuda_arch"] or compute_capability
        if not cuda_arch:
            raise RuntimeError("CUDA architecture could not be determined")
        if cuda_arch.startswith("sm_"):
            cuda_arch = cuda_arch.replace("sm_", "")
        if compute_capability and int(compute_capability) < int(cuda_arch):
            raise RuntimeError(
                f"CUDA architecture {cuda_arch} exceeds compute capability {compute_capability}"
            )
    else:
        cuda_arch = ""

    gt_include_path = gt_config.build_settings["gt_include_path"]

    import os

    extra_compile_args = dict(
        cxx=[
            "-std=c++17",
            "-ftemplate-depth={}".format(gt_config.build_settings["cpp_template_depth"]),
            "-fvisibility=hidden",
            "-fPIC",
            "-isystem{}".format(gt_include_path),
            "-isystem{}".format(gt_config.build_settings["boost_include_path"]),
            "-DBOOST_PP_VARIADICS",
            *extra_compile_args_from_config["cxx"],
        ],
        nvcc=[
            "-std=c++17",
            "-ftemplate-depth={}".format(gt_config.build_settings["cpp_template_depth"]),
            "-arch=sm_{}".format(cuda_arch),
            "-isystem={}".format(gt_include_path),
            "-isystem={}".format(gt_config.build_settings["boost_include_path"]),
            "-DBOOST_PP_VARIADICS",
            "-DBOOST_OPTIONAL_CONFIG_USE_OLD_IMPLEMENTATION_OF_OPTIONAL",
            "-DBOOST_OPTIONAL_USE_OLD_DEFINITION_OF_NONE",
            "--expt-relaxed-constexpr",
            "--compiler-options",
            "-fvisibility=hidden",
            "--compiler-options",
            "-fPIC",
            *extra_compile_args_from_config["nvcc"],
        ],
    )
    extra_link_args = gt_config.build_settings["extra_link_args"]

    mode_flags = ["-O0", "-ggdb"] if debug_mode else ["-O3", "-DNDEBUG"]
    extra_compile_args["cxx"].extend(mode_flags)
    extra_compile_args["nvcc"].extend(mode_flags)
    extra_link_args.extend(mode_flags)

    if dace_path := get_dace_module_path():
        extra_compile_args["cxx"].append(
            "-isystem{}".format(os.path.join(dace_path, "runtime/include"))
        )
        extra_compile_args["nvcc"].append(
            "-isystem={}".format(os.path.join(dace_path, "runtime/include"))
        )

    if add_profile_info:
        profile_flags = ["-pg"]
        extra_compile_args["cxx"].extend(profile_flags)
        extra_compile_args["nvcc"].extend(profile_flags)
        extra_link_args.extend(profile_flags)

    if uses_cuda:
        build_opts = dict(
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
        )
    else:
        build_opts = dict(
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args["cxx"],
            extra_link_args=extra_link_args,
        )

    if uses_openmp:
        cpp_flags = gt_config.build_settings["openmp_cppflags"]
        if uses_cuda:
            cuda_flags = []
            for cpp_flag in cpp_flags:
                cuda_flags.extend(["--compiler-options", cpp_flag])
            build_opts["extra_compile_args"]["nvcc"].extend(cuda_flags)
        elif cpp_flags:
            build_opts["extra_compile_args"].extend(cpp_flags)

        ld_flags = gt_config.build_settings["openmp_ldflags"]
        if ld_flags:
            build_opts["extra_link_args"].extend(ld_flags)

    return build_opts


# The following tells mypy to accept unpacking kwargs
@overload
def build_pybind_ext(
    name: str,
    sources: list,
    build_path: str,
    target_path: str,
    **kwargs: str,
) -> Tuple[str, str]:
    pass


def build_pybind_ext(
    name: str,
    sources: list,
    build_path: str,
    target_path: str,
    *,
    include_dirs: Optional[List[str]] = None,
    library_dirs: Optional[List[str]] = None,
    libraries: Optional[List[str]] = None,
    extra_compile_args: Optional[Union[List[str], Dict[str, List[str]]]] = None,
    extra_link_args: Optional[List[str]] = None,
    build_ext_class: Type = None,
    verbose: bool = False,
    clean: bool = False,
) -> Tuple[str, str]:

    # Hack to remove warning about "-Wstrict-prototypes" not having effect in C++
    replaced_flags_backup = copy.deepcopy(distutils.sysconfig._config_vars)
    _clean_build_flags(distutils.sysconfig._config_vars)

    include_dirs = include_dirs or []
    library_dirs = library_dirs or []
    libraries = libraries or []
    extra_compile_args = extra_compile_args or []
    extra_link_args = extra_link_args or []

    # Build extension module
    py_extension = setuptools.Extension(
        name,
        sources,
        include_dirs=[pybind11.get_include(), pybind11.get_include(user=True), *include_dirs],
        library_dirs=[*library_dirs],
        libraries=[*libraries],
        language="c++",
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )

    setuptools_args = dict(
        name=name,
        ext_modules=[py_extension],
        script_args=[
            "build_ext",
            # "--parallel={}".format(gt_config.build_settings["parallel_jobs"]),
            "--build-temp={}".format(build_path),
            "--build-lib={}".format(build_path),
            "--force",
        ],
    )
    if build_ext_class is not None:
        setuptools_args["cmdclass"] = {"build_ext": build_ext_class}

    if verbose:
        setuptools_args["script_args"].append("-v")
        setuptools.setup(**setuptools_args)
    else:
        setuptools_args["script_args"].append("-q")
        io_out, io_err = io.StringIO(), io.StringIO()
        with contextlib.redirect_stdout(io_out), contextlib.redirect_stderr(io_err):
            setuptools.setup(**setuptools_args)

    # Copy extension in target path
    module_name = py_extension._full_name
    file_path = py_extension._file_name
    src_path = os.path.join(build_path, file_path)
    dest_path = os.path.join(target_path, os.path.basename(file_path))
    os.makedirs(os.path.dirname(dest_path), exist_ok=True)
    distutils.file_util.copy_file(src_path, dest_path, verbose=verbose)

    # Final cleaning
    if clean:
        shutil.rmtree(build_path)

    # Restore original distutils flag config to not break functionality with "-Wstrict-prototypes"-hack for other
    # tools using distutils.
    for key, value in replaced_flags_backup.items():
        distutils.sysconfig._config_vars[key] = value

    return module_name, dest_path


# The following tells mypy to accept unpacking kwargs
@overload
def build_pybind_cuda_ext(
    name: str,
    sources: list,
    build_path: str,
    target_path: str,
    **kwargs: str,
) -> Tuple[str, str]:
    pass


def build_pybind_cuda_ext(
    name: str,
    sources: list,
    build_path: str,
    target_path: str,
    *,
    include_dirs: Optional[List[str]] = None,
    library_dirs: Optional[List[str]] = None,
    libraries: Optional[List[str]] = None,
    extra_compile_args: Optional[Union[List[str], Dict[str, List[str]]]] = None,
    extra_link_args: Optional[List[str]] = None,
    verbose: bool = False,
    clean: bool = False,
) -> Tuple[str, str]:

    include_dirs = include_dirs or []
    include_dirs = [*include_dirs, gt_config.build_settings["cuda_include_path"]]
    library_dirs = library_dirs or []
    library_dirs = [*library_dirs, gt_config.build_settings["cuda_library_path"]]
    libraries = libraries or []
    libraries = [*libraries, "cudart"]
    extra_compile_args = extra_compile_args or []

    return build_pybind_ext(
        name,
        sources,
        build_path,
        target_path,
        verbose=verbose,
        clean=clean,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        libraries=libraries,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        build_ext_class=CUDABuildExtension,
    )


def _clean_build_flags(config_vars: Dict[str, str]) -> None:
    for key, value in config_vars.items():
        if type(value) == str:
            value = " " + value + " "
            for s in value.split(" "):
                if (
                    s in ["-Wstrict-prototypes", "-DNDEBUG", "-pg"]
                    or s.startswith("-O")
                    or s.startswith("-g")
                ):
                    value = value.replace(" " + s + " ", " ")
            config_vars[key] = " ".join(value.split())


class CUDABuildExtension(build_ext, object):
    # Refs:
    #   - https://github.com/pytorch/pytorch/torch/utils/cpp_extension.py
    #   - https://github.com/rmcgibbo/npcuda-example/blob/master/cython/setup.py
    #
    def build_extensions(self) -> None:
        # Register .cu  source extensions
        self.compiler.src_extensions.append(".cu")

        # Save references to the original methods
        original_compile = self.compiler._compile

        def nvcc_compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
            original_compiler_so = self.compiler.compiler_so
            # Copy before we make any modifications.
            cflags = copy.deepcopy(extra_postargs)
            try:
                if os.path.splitext(src)[-1] == ".cu":
                    nvcc_exec = os.path.join(gt_config.build_settings["cuda_bin_path"], "nvcc")
                    self.compiler.set_executable("compiler_so", [nvcc_exec])
                    if isinstance(cflags, dict):
                        cflags = cflags["nvcc"]
                elif isinstance(cflags, dict):
                    cflags = cflags["cxx"]

                original_compile(obj, src, ext, cc_args, cflags, pp_opts)
            finally:
                # Put the original compiler back in place.
                self.compiler.set_executable("compiler_so", original_compiler_so)

        self.compiler._compile = nvcc_compile
        build_ext.build_extensions(self)
        self.compiler._compile = original_compile



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import copy as cp
from typing import Any, Dict, Optional, Set, Type, Union, cast

from gt4py import eve
from gt4py.cartesian.gtc import oir


class MaskCollector(eve.NodeVisitor):
    """Collects the boolean expressions defining mask statements that are boolean fields."""

    def visit_AssignStmt(
        self,
        node: oir.AssignStmt,
        *,
        masks_to_inline: Dict[str, oir.Expr],
    ) -> None:
        if node.left.name in masks_to_inline:
            assert masks_to_inline[node.left.name] is None
            masks_to_inline[node.left.name] = node.right

    def visit_MaskStmt(
        self, node: oir.MaskStmt, *, masks_to_inline: Dict[str, oir.Expr], **kwargs: Any
    ) -> None:
        if isinstance(node.mask, oir.FieldAccess) and node.mask.name in masks_to_inline:
            # Find all reads in condition
            condition_reads: Set[str] = (
                masks_to_inline[node.mask.name]
                .walk_values()
                .if_isinstance(oir.FieldAccess, oir.ScalarAccess)
                .getattr("name")
                .to_set()
            )
            # Find all writes in body
            body_writes: Set[str] = {
                child.left.name for child in node.body if isinstance(child, oir.AssignStmt)
            }
            # Do not inline the mask if there is an intersection
            if condition_reads.intersection(body_writes):
                masks_to_inline.pop(node.mask.name)

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> Dict[str, oir.Expr]:
        masks_to_inline: Dict[str, Optional[oir.Expr]] = {
            mask_stmt.mask.name: None
            for mask_stmt in node.walk_values()
            .if_isinstance(oir.MaskStmt)
            .filter(lambda stmt: isinstance(stmt.mask, oir.FieldAccess))
        }
        self.visit(node.vertical_loops, masks_to_inline=masks_to_inline, **kwargs)
        assert all(value is not None for value in masks_to_inline.values())
        return cast(Dict[str, oir.Expr], masks_to_inline)


class MaskInlining(eve.NodeTranslator):
    """Inlines mask statements that are boolean mask fields with the expression that generates the field.

    Preconditions: Mask statements exist as boolean temporary field accesses.
    Postcondition: Mask statements exist as boolean expressions.
    """

    def visit_FieldAccess(
        self,
        node: oir.FieldAccess,
        *,
        masks_to_inline: Dict[str, oir.Expr],
        **kwargs: Any,
    ) -> oir.Expr:
        if node.name in masks_to_inline:
            return cp.copy(masks_to_inline[node.name])
        return self.generic_visit(node, masks_to_inline=masks_to_inline, **kwargs)

    def visit_AssignStmt(
        self,
        node: oir.AssignStmt,
        *,
        masks_to_inline: Dict[str, oir.Expr],
        **kwargs: Any,
    ) -> Union[oir.AssignStmt, Type[eve.NOTHING]]:
        if node.left.name in masks_to_inline:
            return eve.NOTHING
        return self.generic_visit(node, masks_to_inline=masks_to_inline, **kwargs)

    def visit_Temporary(self, node: oir.Temporary, *, masks_to_inline, **kwargs):
        return node if node.name not in masks_to_inline else eve.NOTHING

    def visit_CacheDesc(self, node: oir.CacheDesc, *, masks_to_inline, **kwargs):
        return node if node.name not in masks_to_inline else eve.NOTHING

    def visit_Stencil(self, node: oir.Stencil, **kwargs):
        return self.generic_visit(node, masks_to_inline=MaskCollector().visit(node), **kwargs)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from . import attrib, meta, text
from .base import (
    NOTHING,
    BaseFrozen,
    BaseSingleton,
    Registry,
    UniqueIdGenerator,
    classmethod_to_function,
    classproperty,
    compose,
    flatten,
    flatten_iter,
    get_member,
    is_collection,
    is_identifier_name,
    is_iterable_of,
    is_mapping_of,
    jsonify,
    make_dir,
    make_local_dir,
    make_module_from_file,
    namespace_from_nested_dict,
    normalize_mapping,
    patch_module,
    restore_module,
    shash,
    shashed_id,
    slugify,
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Dict, List, Optional, Tuple, Union

from gt4py import eve
from gt4py.cartesian.gtc import common, oir, utils
from gt4py.cartesian.gtc.definitions import Extent
from gt4py.cartesian.gtc.passes.horizontal_masks import compute_relative_mask
from gt4py.cartesian.gtc.passes.oir_optimizations.utils import compute_extents

from . import npir


class OirToNpir(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    """Lower from optimizable IR (OIR) to numpy IR (NPIR)."""

    # --- Decls ---
    def visit_FieldDecl(
        self, node: oir.FieldDecl, *, field_extents: Dict[str, Extent], **kwargs: Any
    ) -> npir.FieldDecl:
        extent = field_extents.get(node.name, Extent.zeros(ndims=2))
        return npir.FieldDecl(
            name=node.name,
            dtype=node.dtype,
            dimensions=node.dimensions,
            data_dims=node.data_dims,
            extent=extent,
        )

    def visit_ScalarDecl(self, node: oir.ScalarDecl, **kwargs: Any) -> npir.ScalarDecl:
        return npir.ScalarDecl(name=node.name, dtype=node.dtype)

    def visit_LocalScalar(self, node: oir.LocalScalar, **kwargs: Any) -> npir.LocalScalarDecl:
        return npir.LocalScalarDecl(name=node.name, dtype=node.dtype)

    def visit_Temporary(
        self, node: oir.Temporary, *, field_extents: Dict[str, Extent], **kwargs: Any
    ) -> npir.TemporaryDecl:
        temp_extent = field_extents[node.name]
        offset = tuple(-ext[0] for ext in temp_extent)
        assert all(off >= 0 for off in offset)
        padding = tuple(ext[1] - ext[0] for ext in temp_extent)
        return npir.TemporaryDecl(
            name=node.name,
            dtype=node.dtype,
            data_dims=node.data_dims,
            offset=offset,
            padding=padding,
        )

    # --- Expressions ---
    def visit_Literal(self, node: oir.Literal, **kwargs: Any) -> npir.ScalarLiteral:
        assert node.kind == common.ExprKind.SCALAR
        return npir.ScalarLiteral(value=node.value, dtype=node.dtype, kind=node.kind)

    def visit_ScalarAccess(
        self, node: oir.ScalarAccess, *, symtable: Dict[str, oir.Decl], **kwargs: Any
    ) -> Union[npir.ParamAccess, npir.LocalScalarAccess]:
        assert node.kind == common.ExprKind.SCALAR
        if isinstance(symtable[node.name], oir.LocalScalar):
            return npir.LocalScalarAccess(name=node.name, dtype=symtable[node.name].dtype)
        else:
            return npir.ParamAccess(name=node.name, dtype=symtable[node.name].dtype)

    def visit_CartesianOffset(
        self, node: common.CartesianOffset, **kwargs: Any
    ) -> Tuple[int, int, int]:
        return node.i, node.j, node.k

    def visit_VariableKOffset(
        self, node: oir.VariableKOffset, **kwargs: Any
    ) -> Tuple[int, int, eve.Node]:
        return 0, 0, npir.VarKOffset(k=self.visit(node.k, **kwargs))

    def visit_FieldAccess(self, node: oir.FieldAccess, **kwargs: Any) -> npir.FieldSlice:
        i_offset, j_offset, k_offset = self.visit(node.offset, **kwargs)
        data_index = [self.visit(index, **kwargs) for index in node.data_index]
        return npir.FieldSlice(
            name=node.name,
            i_offset=i_offset,
            j_offset=j_offset,
            k_offset=k_offset,
            data_index=data_index,
            dtype=node.dtype,
        )

    def visit_UnaryOp(self, node: oir.UnaryOp, **kwargs: Any) -> npir.VectorUnaryOp:
        return npir.VectorUnaryOp(op=node.op, expr=self.visit(node.expr, **kwargs))

    def visit_BinaryOp(
        self, node: oir.BinaryOp, **kwargs: Any
    ) -> Union[npir.VectorArithmetic, npir.VectorLogic]:
        args = dict(
            op=node.op,
            left=self.visit(node.left, **kwargs),
            right=self.visit(node.right, **kwargs),
        )
        if isinstance(node.op, common.LogicalOperator):
            return npir.VectorLogic(**args)
        else:
            return npir.VectorArithmetic(**args)

    def visit_TernaryOp(self, node: oir.TernaryOp, **kwargs: Any) -> npir.VectorTernaryOp:
        return npir.VectorTernaryOp(
            cond=self.visit(node.cond, **kwargs),
            true_expr=self.visit(node.true_expr, **kwargs),
            false_expr=self.visit(node.false_expr, **kwargs),
        )

    def visit_Cast(self, node: oir.Cast, **kwargs: Any) -> Union[npir.VectorCast, npir.ScalarCast]:
        expr = self.visit(node.expr, **kwargs)
        args = {"dtype": node.dtype, "expr": expr}
        return (
            npir.VectorCast(**args)
            if expr.kind == common.ExprKind.FIELD
            else npir.ScalarCast(**args)
        )

    def visit_NativeFuncCall(self, node: oir.NativeFuncCall, **kwargs: Any) -> npir.NativeFuncCall:
        return npir.NativeFuncCall(
            func=self.visit(node.func, **kwargs), args=self.visit(node.args, **kwargs)
        )

    # --- Statements ---
    def visit_MaskStmt(
        self,
        node: oir.MaskStmt,
        *,
        mask: Optional[npir.Expr] = None,
        **kwargs: Any,
    ) -> List[npir.Stmt]:
        mask_expr = self.visit(node.mask, **kwargs)
        if mask:
            mask_expr = npir.VectorLogic(op=common.LogicalOperator.AND, left=mask, right=mask_expr)

        return utils.flatten_list(self.visit(node.body, mask=mask_expr, **kwargs))

    def visit_AssignStmt(
        self,
        node: oir.AssignStmt,
        *,
        mask: Optional[npir.Expr] = None,
        horizontal_mask: Optional[npir.HorizontalMask] = None,
        **kwargs: Any,
    ) -> npir.VectorAssign:
        left = self.visit(node.left, **kwargs)
        right = self.visit(node.right, **kwargs)

        if right.kind == common.ExprKind.SCALAR:
            right = npir.Broadcast(expr=right)

        if mask is not None:
            right = npir.VectorTernaryOp(cond=mask, true_expr=right, false_expr=left)

        return npir.VectorAssign(left=left, right=right, horizontal_mask=horizontal_mask)

    def visit_While(
        self, node: oir.While, *, mask: Optional[npir.Expr] = None, **kwargs: Any
    ) -> npir.While:
        cond = self.visit(node.cond, mask=mask, **kwargs)
        if mask:
            mask = npir.VectorLogic(op=common.LogicalOperator.AND, left=mask, right=cond)
        else:
            mask = cond
        return npir.While(
            cond=cond, body=utils.flatten_list(self.visit(node.body, mask=mask, **kwargs))
        )

    def visit_HorizontalRestriction(
        self, node: oir.HorizontalRestriction, *, extent: Extent, **kwargs: Any
    ) -> Any:
        mask = compute_relative_mask(extent, node.mask)
        if mask is None:
            return eve.NOTHING

        horizontal_mask = npir.HorizontalMask(i=mask[0], j=mask[1])

        return utils.flatten_list(self.visit(node.body, horizontal_mask=horizontal_mask, **kwargs))

    # --- Control Flow ---
    def visit_HorizontalExecution(
        self,
        node: oir.HorizontalExecution,
        *,
        block_extents: Optional[Dict[int, Extent]] = None,
        **kwargs: Any,
    ) -> npir.HorizontalBlock:
        if block_extents:
            extent = block_extents[id(node)]
        else:
            extent = Extent.zeros(ndims=2)

        stmts = utils.flatten_list(self.visit(node.body, extent=extent, **kwargs))
        return npir.HorizontalBlock(
            body=stmts, extent=extent, declarations=self.visit(node.declarations, **kwargs)
        )

    def visit_VerticalLoopSection(
        self, node: oir.VerticalLoopSection, *, loop_order: common.LoopOrder, **kwargs: Any
    ) -> npir.VerticalPass:
        return npir.VerticalPass(
            body=self.visit(node.horizontal_executions, **kwargs),
            lower=node.interval.start,
            upper=node.interval.end,
            direction=loop_order,
        )

    def visit_VerticalLoop(self, node: oir.VerticalLoop, **kwargs: Any) -> List[npir.VerticalPass]:
        return self.visit(node.sections, loop_order=node.loop_order, **kwargs)

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> npir.Computation:
        field_extents, block_extents = compute_extents(node)

        arguments = [decl.name for decl in node.params]
        param_decls = [
            self.visit(decl, **kwargs) for decl in node.params if isinstance(decl, oir.ScalarDecl)
        ]
        api_field_decls = [
            self.visit(decl, field_extents=field_extents)
            for decl in node.params
            if isinstance(decl, oir.FieldDecl)
        ]
        temp_decls = [
            self.visit(decl, field_extents=field_extents, **kwargs) for decl in node.declarations
        ]

        vertical_passes = utils.flatten_list(
            self.visit(
                node.vertical_loops,
                block_extents=block_extents,
                **kwargs,
            )
        )

        return npir.Computation(
            arguments=arguments,
            api_field_decls=api_field_decls,
            param_decls=param_decls,
            temp_decls=temp_decls,
            vertical_passes=vertical_passes,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Visitor classes to work with IR trees."""


from __future__ import annotations

import collections.abc
import copy

from . import concepts, trees
from .extended_typing import Any
from .type_definitions import NOTHING


class NodeVisitor:
    """Simple node visitor class based on :class:`ast.NodeVisitor`.

    A NodeVisitor instance walks a node tree and calls a visitor
    function for every item found. This class is meant to be subclassed,
    with the subclass adding visitor methods.

    Visitor functions for tree elements are named with a standard
    pattern: ``visit_`` + class name of the node. Thus, the visitor
    function for a ``BinOpExpr`` node class should be called ``visit_BinOpExpr``.
    If no visitor function exists for a specific node, the dispatcher
    mechanism looks for the visitor of each one of its parent classes
    in the order define by the class' ``__mro__`` attribute. Finally,
    if no visitor function has been found, the ``generic_visit`` visitor
    is used instead. A concise summary of the steps followed to find an
    appropriate visitor function is:

        1. A ``self.visit_NODE_CLASS_NAME()`` method where `NODE_CLASS_NAME`
           matches ``type(node).__name__``.
        2. A ``self.visit_NODE_BASE_CLASS_NAME()`` method where
           `NODE_BASE_CLASS_NAME` matches ``base.__name__``, and `base` is
           one of the node base classes (evaluated following the order
           given in ``type(node).__mro__``).
        3. ``self.generic_visit()``.

    This dispatching mechanism is implemented in the main :meth:`visit`
    method and can be overriden in subclasses. Therefore, a simple way to extend
    the behavior of a visitor is by inheriting from lightweight `trait` classes
    with a custom ``visit()`` method, which wraps the call to the superclass'
    ``visit()`` and adds extra pre and post visit logic. Check :mod:`eve.traits`
    for further information.

    Note that return values are not forwarded to the caller in the default
    :meth:`generic_visit` implementation. If you want to return a value from
    a nested node in the tree, make sure all the intermediate nodes explicitly
    return children values.

    The recommended idiom to use and define NodeVisitors can be summarized as:

        * Inside visitor functions:

            + call ``self.visit(child_node)`` to visit children.
            + call ``self.generic_visit(node)`` to continue the tree
              traversal in the usual way.

        * Define an ``apply()`` `classmethod` as a shortcut to create an
          instance and start the visit::

                class Visitor(NodeVisitor)
                    @classmethod
                    def apply(cls, tree, init_var, foo, bar=5, **kwargs):
                        instance = cls(init_var)
                        return instance(tree, foo=foo, bar=bar, **kwargs)

                    ...

                result = Visitor.apply(...)

        * If the visitor has internal state, make sure visitor instances
          are never reused or clean up the state at the end.

    Notes:
        If you want to apply changes to nodes during the traversal,
        use the :class:`NodeMutator` subclass, which handles correctly
        structural modifications of the visited tree.

    """

    def visit(self, node: concepts.RootNode, **kwargs: Any) -> Any:
        visitor = self.generic_visit

        class_name = node.__class__.__name__
        if "__" in class_name:
            # For concretized data model classes, use the generic name
            class_name = class_name[: class_name.find("__")]
        method_name = "visit_" + class_name

        if hasattr(self, method_name):
            visitor = getattr(self, method_name)
        elif isinstance(node, concepts.Node):
            for node_class in node.__class__.__mro__[1:]:
                class_name = node_class.__name__
                if "__" in class_name:
                    class_name = class_name[: class_name.find("__")]
                method_name = "visit_" + class_name

                if hasattr(self, method_name):
                    visitor = getattr(self, method_name)
                    break

                if node_class is concepts.Node:
                    break

        return visitor(node, **kwargs)

    def generic_visit(self, node: concepts.RootNode, **kwargs: Any) -> Any:
        for child in trees.iter_children_values(node):
            self.visit(child, **kwargs)

        return None


class NodeTranslator(NodeVisitor):
    """Special `NodeVisitor` to translate nodes and trees.

    A NodeTranslator instance will walk the tree exactly as a regular
    :class:`NodeVisitor` while building an output tree using the return
    values of the visitor methods. If the return value is :obj:`eve.NOTHING`,
    the node will be removed from its location in the output tree,
    otherwise it will be replaced with this new value. The default visitor
    method (:meth:`generic_visit`) returns a `deepcopy` of the original
    node.

    Keep in mind that if the node you're operating on has child nodes
    you must either transform the child nodes yourself or call the
    :meth:`generic_visit` method for the node first.

    Usually you use a NodeTranslator like this::

       output_node = YourTranslator.apply(input_node)

    Notes:
        Check :class:`NodeVisitor` documentation for more details.

    """

    def generic_visit(self, node: concepts.RootNode, **kwargs: Any) -> Any:

        memo = kwargs.get("__memo__", None)

        if isinstance(node, concepts.Node):
            new_node = node.__class__(  # type: ignore
                **{
                    name: new_child
                    for name, child in node.iter_children_items()
                    if (new_child := self.visit(child, **kwargs)) is not NOTHING
                },
            )
            return new_node

        if isinstance(node, (list, tuple, set, collections.abc.Set)) or (
            isinstance(node, collections.abc.Sequence) and not isinstance(node, (str, bytes))
        ):
            # Sequence or set: create a new container instance with the new values
            return node.__class__(  # type: ignore
                new_child
                for child in trees.iter_children_values(node)
                if (new_child := self.visit(child, **kwargs)) is not NOTHING
            )

        if isinstance(node, (dict, collections.abc.Mapping)):
            # Mapping: create a new mapping instance with the new values
            return node.__class__(  # type: ignore[call-arg]
                {
                    name: new_child
                    for name, child in trees.iter_children_items(node)
                    if (new_child := self.visit(child, **kwargs)) is not NOTHING
                }
            )

        return copy.deepcopy(node, memo=memo)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Typing definitions working across different Python versions (via `typing_extensions`).

Definitions in 'typing_extensions' take priority over those in 'typing'.
"""

from __future__ import annotations

import abc as _abc
import array as _array
import collections.abc as _collections_abc
import dataclasses as _dataclasses
import enum as _enum
import functools as _functools
import inspect as _inspect
import mmap as _mmap
import pickle as _pickle
import sys as _sys
import types as _types
import typing as _typing
from typing import *  # noqa: F403
from typing import overload  # Only needed to avoid false flake8 errors

import typing_extensions as _typing_extensions
from typing_extensions import *  # type: ignore[misc]  # noqa: F403


if _sys.version_info >= (3, 9):
    # Standard library already supports PEP 585 (Type Hinting Generics In Standard Collections)
    from builtins import (  # type: ignore[misc]  # isort:skip
        tuple as Tuple,
        list as List,
        dict as Dict,
        set as Set,
        frozenset as FrozenSet,
        type as Type,
    )
    from collections import (  # isort:skip
        ChainMap as ChainMap,
        Counter as Counter,
        OrderedDict as OrderedDict,
        defaultdict as defaultdict,
        deque as deque,
    )
    from collections.abc import (  # isort:skip
        AsyncGenerator as AsyncGenerator,
        AsyncIterable as AsyncIterable,
        AsyncIterator as AsyncIterator,
        Awaitable as Awaitable,
        ByteString as ByteString,
        Callable as Callable,
        Collection as Collection,
        Container as Container,
        Coroutine as Coroutine,
        Generator as Generator,
        ItemsView as ItemsView,
        Iterable as Iterable,
        Iterator as Iterator,
        KeysView as KeysView,
        Mapping as Mapping,
        MappingView as MappingView,
        MutableMapping as MutableMapping,
        MutableSequence as MutableSequence,
        MutableSet as MutableSet,
        Reversible as Reversible,
        Sequence as Sequence,
    )
    from collections.abc import Set as AbstractSet  # isort:skip
    from collections.abc import ValuesView as ValuesView  # isort:skip
    from contextlib import (  # isort:skip
        AbstractAsyncContextManager as AsyncContextManager,
    )
    from contextlib import AbstractContextManager as ContextManager  # isort:skip
    from re import Match as Match, Pattern as Pattern  # isort:skip


# These fallbacks are useful for public symbols not exported by default.
# Again, definitions in 'typing_extensions' take priority over those in 'typing'
def __getattr__(name: str) -> Any:
    import sys

    import typing_extensions

    result = SENTINEL = object()
    if not (name.startswith("__") and name.endswith("__")):
        result = getattr(typing_extensions, name, SENTINEL)
        if result is SENTINEL:
            import typing

            result = getattr(typing, name, SENTINEL)

    if result is SENTINEL:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

    setattr(sys.modules[__name__], name, result)  # cache result

    return result


def __dir__() -> List[str]:
    if not hasattr(self_func := (globals()["__dir__"]), "__cached_dir"):
        import typing

        import typing_extensions

        orig_dir = typing.__dir__()
        self_func.__cached_dir = [*orig_dir] + [
            name for name in typing_extensions.__dir__() if name not in orig_dir
        ]

    return self_func.__cached_dir


_T = TypeVar("_T")

# -- Common type aliases --
NoArgsCallable = Callable[[], Any]


# -- Typing annotations --
if _sys.version_info >= (3, 9):
    SolvedTypeAnnotation = Union[
        Type,
        _typing._SpecialForm,
        _types.GenericAlias,  # type: ignore[name-defined]  # Python 3.8 does not include `_types.GenericAlias`
        _typing._BaseGenericAlias,  # type: ignore[name-defined]  # _BaseGenericAlias is not exported in stub
    ]
else:
    SolvedTypeAnnotation = Union[  # type: ignore[misc]  # mypy consider this assignment a redefinition
        Type,
        _typing._SpecialForm,
        _typing._GenericAlias,  # type: ignore[attr-defined]  # _GenericAlias is not exported in stub
    ]

TypeAnnotation = Union[ForwardRef, SolvedTypeAnnotation]
SourceTypeAnnotation = Union[str, TypeAnnotation]

StdGenericAliasType: Final[Type] = type(List[int])

if _sys.version_info >= (3, 9):
    if TYPE_CHECKING:
        StdGenericAlias: TypeAlias = _types.GenericAlias  # type: ignore[name-defined,attr-defined]  # Python 3.8 does not include `_types.GenericAlias`

_TypingSpecialFormType: Final[Type] = _typing._SpecialForm
_TypingGenericAliasType: Final[Type] = (
    _typing._BaseGenericAlias if _sys.version_info >= (3, 9) else _typing._GenericAlias  # type: ignore[attr-defined]  # _BaseGenericAlias / _GenericAlias are not exported in stub
)


# -- Standard Python protocols --
_C = TypeVar("_C")
_V = TypeVar("_V")


class NonDataDescriptor(Protocol[_C, _V]):
    """Typing protocol for non-data descriptor classes.

    See https://docs.python.org/3/howto/descriptor.html for further information.
    """

    @overload
    def __get__(
        self, _instance: Literal[None], _owner_type: Optional[Type[_C]] = None
    ) -> NonDataDescriptor[_C, _V]:
        ...

    @overload
    def __get__(  # noqa: F811  # redefinion of unused member
        self, _instance: _C, _owner_type: Optional[Type[_C]] = None
    ) -> _V:
        ...

    def __get__(  # noqa: F811  # redefinion of unused member
        self, _instance: Optional[_C], _owner_type: Optional[Type[_C]] = None
    ) -> _V | NonDataDescriptor[_C, _V]:
        ...


class DataDescriptor(NonDataDescriptor[_C, _V], Protocol):
    """Typing protocol for data descriptor classes.

    See https://docs.python.org/3/howto/descriptor.html for further information.
    """

    def __set__(self, _instance: _C, _value: _V) -> None:
        ...

    def __delete__(self, _instance: _C) -> None:
        ...


# -- Based on typeshed definitions --
ReadOnlyBuffer: TypeAlias = Union[bytes, SupportsBytes]
WriteableBuffer: TypeAlias = Union[
    bytearray, memoryview, _array.array, _mmap.mmap, _pickle.PickleBuffer
]
ReadableBuffer: TypeAlias = Union[ReadOnlyBuffer, WriteableBuffer]


class HashlibAlgorithm(Protocol):
    """Used in the hashlib module of the standard library."""

    digest_size: int
    block_size: int
    name: str

    def __init__(self, data: ReadableBuffer = ...) -> None:
        ...

    def copy(self) -> HashlibAlgorithm:
        ...

    def update(self, data: ReadableBuffer) -> None:
        ...

    def digest(self) -> bytes:
        ...

    def hexdigest(self) -> str:
        ...


# -- Third party protocols --
class DevToolsPrettyPrintable(Protocol):
    """Used by python-devtools (https://python-devtools.helpmanual.io/)."""

    def __pretty__(self, fmt: Callable[[Any], Any], **kwargs: Any) -> Generator[Any, None, None]:
        ...


# -- Added functionality --
class NonProtocolABCMeta(_typing._ProtocolMeta):
    """Subclass of :cls:`typing.Protocol`'s metaclass doing instance and subclass checks as ABCMeta."""

    __instancecheck__ = _abc.ABCMeta.__instancecheck__  # type: ignore[assignment]
    __subclasshook__ = None


_ProtoT = TypeVar("_ProtoT", bound=_abc.ABCMeta)


@overload
def extended_runtime_checkable(
    *,
    instance_check_shortcut: bool = True,
    subclass_check_with_data_members: bool = False,
) -> Callable[[_ProtoT], _ProtoT]:
    ...


@overload
def extended_runtime_checkable(
    maybe_cls: _ProtoT,
    *,
    instance_check_shortcut: bool = True,
    subclass_check_with_data_members: bool = False,
) -> _ProtoT:
    ...


def extended_runtime_checkable(  # noqa: C901  # too complex but unavoidable
    maybe_cls: Optional[_ProtoT] = None,
    *,
    instance_check_shortcut: bool = True,
    subclass_check_with_data_members: bool = False,
) -> _ProtoT | Callable[[_ProtoT], _ProtoT]:
    """Emulates :func:`typing.runtime_checkable` with optional performance shortcuts.

    If all optional shortcuts are set to ``False``, it behaves exactly
    as :func:`typing.runtime_checkable`.

    Keyword Arguments:
        instance_check_shortcut: instance checks only use the instance type
            instead of checking the instance data for members added at runtime.
        subclass_check_with_data_members: subclass checks also work for
            protocols with data members.
    """

    def _decorator(cls: _ProtoT) -> _ProtoT:
        cls = _typing.runtime_checkable(cls)
        if not (instance_check_shortcut or subclass_check_with_data_members):
            return cls

        if instance_check_shortcut:
            # Monkey patch the decorated protocol class using our custom
            # metaclass, which assumes that no data members have been
            # added at runtime and therefore the expensive instance members
            # checks can be replaced by (cached) tests with class members
            cls.__class__ = NonProtocolABCMeta  # type: ignore[assignment]

        if subclass_check_with_data_members:
            assert "__subclasshook__" in cls.__dict__
            if cls.__subclasshook__.__module__ not in (  # type: ignore[attr-defined]
                "typing",
                "typing_extensions",
                "extended_typing",
            ):
                raise TypeError(
                    "Cannot use 'subclass_check_with_data_members' with custom '__subclasshook__' definitions."
                )

            _allow_reckless_class_checks = getattr(
                _typing,
                "_allow_reckless_class_checks"
                if hasattr(_typing, "_allow_reckless_class_checks")
                else "_allow_reckless_class_cheks",  # There is a typo in 3.8 and 3.9
            )

            _get_protocol_attrs = (
                _typing._get_protocol_attrs  # type: ignore[attr-defined]  # private member
            )
            _is_callable_members_only = (
                _typing._is_callable_members_only  # type: ignore[attr-defined]  # private member
            )

            # Define a patched version of the proto hook which ignores
            # __is_callable_members_only() result at certain points
            def _patched_proto_hook(other):  # type: ignore[no-untyped-def]
                if not cls.__dict__.get("_is_protocol", False):
                    return NotImplemented

                # First, perform various sanity checks.
                if not getattr(cls, "_is_runtime_protocol", False):
                    if _allow_reckless_class_checks():
                        return NotImplemented
                    raise TypeError(
                        "Instance and class checks can only be used with"
                        " @runtime_checkable protocols"
                    )
                if not _is_callable_members_only(cls) and _allow_reckless_class_checks():
                    return NotImplemented
                    # PATCHED: a TypeError should be raised here if not
                    # `allow_reckless_class_checks()` but we ignored in
                    # this patched version`
                if not isinstance(other, type):
                    # Same error message as for issubclass(1, int).
                    raise TypeError("issubclass() arg 1 must be a class")

                # Second, perform the actual structural compatibility check.
                for attr in _get_protocol_attrs(cls):
                    for base in other.__mro__:
                        # Check if the members appears in the class dictionary...
                        if callable(getattr(cls, attr, None)):  # Method member
                            if attr in base.__dict__:
                                if base.__dict__[attr] is None:
                                    return NotImplemented
                                break
                        elif attr in base.__dict__ or (  # Data member
                            base_annotations := getattr(base, "__annotations__", {})
                            and isinstance(base_annotations, _collections_abc.Mapping)
                            and attr in base_annotations
                        ):
                            break

                        # ...or in annotations, if it is a sub-protocol.
                        base_annotations = getattr(base, "__annotations__", {})
                        if (
                            isinstance(base_annotations, _collections_abc.Mapping)
                            and attr in base_annotations
                            and issubclass(other, Generic)
                            and other._is_protocol
                        ):
                            break
                    else:
                        return NotImplemented
                return True

            cls.__subclasshook__ = _patched_proto_hook  # type: ignore[attr-defined]

        return cls

    return _decorator(maybe_cls) if maybe_cls is not None else _decorator


_ArtefactTypes: tuple[type, ...] = tuple()
if _sys.version_info >= (3, 9):
    _ArtefactTypes = (_types.GenericAlias,)  # type: ignore[attr-defined]  # GenericAlias only from >= 3.8

    # `Any` is a class since Python 3.11
    if isinstance(_typing.Any, type):  # Python >= 3.11
        _ArtefactTypes = (*_ArtefactTypes, _typing.Any)

# `Any` is a class since typing_extensions >= 4.4
if (typing_exts_any := getattr(_typing_extensions, "Any", None)) is not _typing.Any and isinstance(
    typing_exts_any, type
):
    _ArtefactTypes = (*_ArtefactTypes, typing_exts_any)


def is_actual_type(obj: Any) -> TypeGuard[Type]:
    """Check if an object has an actual type and instead of a typing artefact like ``GenericAlias`` or ``Any``.

    This is needed because since Python 3.9:
        ``isinstance(types.GenericAlias(),  type) is True``
    and since Python 3.11:
        ``isinstance(typing.Any,  type) is True``
    """
    return isinstance(obj, type) and type(obj) not in _ArtefactTypes


if hasattr(_typing_extensions, "Any") and _typing.Any is not _typing_extensions.Any:  # type: ignore[attr-defined] # _typing_extensions.Any only from >= 4.4
    # When using Python < 3.11 and typing_extensions >= 4.4 there are
    # two different implementations of `Any`

    def is_Any(obj: Any) -> bool:
        return obj is _typing.Any or obj is _typing_extensions.Any  # type: ignore[attr-defined] # _typing_extensions.Any only from >= 4.4

else:

    def is_Any(obj: Any) -> bool:
        return obj is _typing.Any


def has_type_parameters(cls: Type) -> bool:
    """Return ``True`` if obj is a generic class with type parameters."""
    return issubclass(cls, Generic) and len(getattr(cls, "__parameters__", [])) > 0  # type: ignore[arg-type]  # Generic not considered as a class


def get_actual_type(obj: _T) -> Type[_T]:
    """Return type of an object (also working for GenericAlias instances which pretend to be an actual type)."""
    return StdGenericAliasType if isinstance(obj, StdGenericAliasType) else type(obj)


def is_type_with_custom_hash(type_: Type) -> bool:
    return type_.__hash__ not in (None, object.__hash__)


class HasCustomHash(Hashable):
    """ABC for types defining a custom hash function."""

    @classmethod
    def __subclasshook__(cls, candidate_cls: type) -> bool:
        return is_type_with_custom_hash(candidate_cls)


def is_value_hashable(obj: Any) -> TypeGuard[HasCustomHash]:
    return isinstance(obj, type) or obj is None or is_type_with_custom_hash(type(obj))


def is_value_hashable_typing(
    type_annotation: TypeAnnotation,
    *,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
) -> bool:
    """Check if a type annotation describes a type hashable by value."""
    if is_actual_type(type_annotation):
        assert not get_args(type_annotation)
        return (
            True
            if type_annotation in (type, type(None))
            else is_type_with_custom_hash(type_annotation)
        )

    if isinstance(type_annotation, TypeVar):
        if type_annotation.__bound__:
            return is_value_hashable_typing(type_annotation.__bound__)
        if type_annotation.__constraints__:
            return all(is_value_hashable_typing(c) for c in type_annotation.__constraints__)
        return False

    if isinstance(type_annotation, ForwardRef):
        return is_value_hashable_typing(
            eval_forward_ref(type_annotation, globalns=globalns, localns=localns)
        )

    if type_annotation is Any:
        return False

    # Generic types
    origin_type = get_origin(type_annotation)
    type_args = get_args(type_annotation)

    if origin_type is Literal:
        return True

    if origin_type is Union:
        return all(is_value_hashable_typing(t) for t in type_args)

    if isinstance(origin_type, type) and is_value_hashable_typing(origin_type):
        return all(is_value_hashable_typing(t) for t in type_args if t != Ellipsis)

    return type_annotation is None


def is_protocol(type_: Type) -> bool:
    """Check if a type is a Protocol definition."""
    return getattr(type_, "_is_protocol", False)


def get_partial_type_hints(
    obj: Union[
        object,
        Callable,
        _types.FunctionType,
        _types.BuiltinFunctionType,
        _types.MethodType,
        _types.ModuleType,
        _types.WrapperDescriptorType,
        _types.MethodWrapperType,
        _types.MethodDescriptorType,
    ],
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    include_extras: bool = False,
) -> Dict[str, Union[Type, ForwardRef]]:
    """Return a dictionary with type hints (using forward refs for undefined names) for a function, method, module or class object.

    For each member type hint in the object a :class:`typing.ForwarRef` instance will be
    returned if some names in the string annotation have not been found. For additional
    information see :func:`typing.get_type_hints`.
    """
    if getattr(obj, "__no_type_check__", None):
        return {}
    if not hasattr(obj, "__annotations__"):
        return get_type_hints(  # type: ignore[call-arg]  # Python 3.8 does not define `include-extras`
            obj, globalns=globalns, localns=localns, include_extras=include_extras
        )

    hints: Dict[str, Union[Type, ForwardRef]] = {}
    annotations = getattr(obj, "__annotations__", {})
    for name, hint in annotations.items():
        obj.__annotations__ = {name: hint}
        try:
            resolved_hints = get_type_hints(  # type: ignore[call-arg]  # Python 3.8 does not define `include-extras`
                obj, globalns=globalns, localns=localns, include_extras=include_extras
            )
            hints.update(resolved_hints)
        except NameError as error:
            if isinstance(hint, str):
                hints[name] = ForwardRef(hint)
            elif isinstance(hint, (ForwardRef, _typing.ForwardRef)):
                hints[name] = hint
            else:
                raise error

    obj.__annotations__ = annotations

    return hints


def eval_forward_ref(
    ref: Union[str, ForwardRef],
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    *,
    include_extras: bool = False,
) -> SolvedTypeAnnotation:
    """Resolve forward references in type annotations.

    Arguments:
        globalns: globals ``dict`` used in the evaluation of the annotations.
        localns: locals ``dict`` used in the evaluation of the annotations.

    Keyword Arguments:
        include_extras: if ``True``, ``Annotated`` hints will preserve the annotation.

    Examples:
        >>> from typing import Dict, Tuple
        >>> print("Result:", eval_forward_ref('Dict[str, Tuple[int, float]]'))
        Result: ...ict[str, ...uple[int, float]]

    """
    actual_type = ForwardRef(ref) if isinstance(ref, str) else ref

    def _f() -> None:
        pass

    _f.__annotations__ = {"ref": actual_type}

    if localns:
        safe_localns = {**localns}
        safe_localns.setdefault("typing", _sys.modules[__name__])
        safe_localns.setdefault("NoneType", type(None))
    else:
        safe_localns = {"typing": _sys.modules[__name__], "NoneType": type(None)}

    actual_type = get_type_hints(_f, globalns, safe_localns, include_extras=include_extras)["ref"]  # type: ignore[call-arg]  # Python 3.8 does not define `include-extras`
    assert not isinstance(actual_type, ForwardRef)

    return actual_type


def _collapse_type_args(*args: Any) -> Tuple[bool, Tuple]:
    if args and all(args[0] == a for a in args[1:]):
        return (True, args)
    else:
        return (False, args)


@final
@_dataclasses.dataclass
class CallableKwargsInfo:
    data: Dict[str, Any]


def infer_type(  # noqa: C901  # function is complex but well organized in independent cases
    value: Any,
    *,
    annotate_callable_kwargs: bool = False,
    none_as_type: bool = True,
) -> TypeAnnotation:
    """Generate a typing definition from a value.

    Keyword Arguments:
        annotate_callable_kwargs: if ``True``, ``Callable``s will be returned as
            a ``Annotated[Callable, CallableKwargsInfo]`` hint, where :class:`CallableKwargsInfo`
            contains the inferred typings for the keyword arguments, if any.
        none_as_type:  if ``True``, ``None`` hints will be transformed to ``type(None)``.

    Examples:
        >>> infer_type(3)
        <class 'int'>

        >>> infer_type((3, "four"))
        tuple[int, str]

        >>> infer_type((3, 4))
        tuple[int, int]

        >>> infer_type(frozenset([1, 2, 3]))
        frozenset[int]

        >>> infer_type({'a': 0, 'b': 1})
        dict[str, int]

        >>> infer_type({'a': 0, 'b': 'B'})
        dict[str, ...Any]

        >>> print("Result:", infer_type(lambda a, b: a + b))
        Result: ...Callable[[...Any, ...Any], ...Any]

        # Note that some patch versions of cpython3.9 show weird behaviors
        >>> def f(a: int, b) -> int: ...
        >>> print("Result:", infer_type(f))
        Result: ...Callable[[...int..., ...Any], int]

        >>> def f(a: int, b) -> int: ...
        >>> print("Result:", infer_type(f))
        Result: ...Callable[..., int]

        >>> print("Result:", infer_type(Dict[int, Union[int, float]]))
        Result: ...ict[int, typing.Union[int, float]]

    For advanced cases, using :func:`functools.singledispatch` with custom hooks
    is a simple way to extend and customize this base implementation.

    Example:
        >>> import functools, numbers
        >>> extended_infer_type = functools.singledispatch(infer_type)
        >>> @extended_infer_type.register(int)
        ... @extended_infer_type.register(float)
        ... @extended_infer_type.register(complex)
        ... def _infer_type_number(value, *, annotate_callable_kwargs: bool = False):
        ...    return numbers.Number
        >>> extended_infer_type(3.4)
        <class 'numbers.Number'>
        >>> infer_type(3.4)
        <class 'float'>

    """
    _infer = _functools.partial(infer_type, annotate_callable_kwargs=annotate_callable_kwargs)

    if isinstance(value, (StdGenericAliasType, _TypingSpecialFormType)):
        return value

    if value in (None, type(None)):
        return type(None) if none_as_type else None

    if isinstance(value, type):
        return Type[value]

    if isinstance(value, tuple):
        _, args = _collapse_type_args(*(_infer(item) for item in value))
        if args:
            return StdGenericAliasType(tuple, args)
        else:
            return StdGenericAliasType(tuple, (Any, ...))

    if isinstance(value, (list, set, frozenset)):
        t: Union[Type[List], Type[Set], Type[FrozenSet]] = type(value)
        unique_type, args = _collapse_type_args(*(_infer(item) for item in value))
        return StdGenericAliasType(t, args[0] if unique_type else Any)

    if isinstance(value, dict):
        unique_key_type, keys = _collapse_type_args(*(_infer(key) for key in value.keys()))
        unique_value_type, values = _collapse_type_args(*(_infer(v) for v in value.values()))
        kt = keys[0] if unique_key_type else Any
        vt = values[0] if unique_value_type else Any
        return StdGenericAliasType(dict, (kt, vt))

    if isinstance(value, _types.FunctionType):
        try:
            annotations = get_type_hints(value)
            return_type = annotations.get("return", Any)

            sig = _inspect.signature(value)
            arg_types: List = []
            kwonly_arg_types: Dict[str, Any] = {}
            for p in sig.parameters.values():
                if p.kind in (
                    _inspect.Parameter.POSITIONAL_ONLY,
                    _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ):
                    arg_types.append(annotations.get(p.name, None) or Any)
                elif p.kind == _inspect.Parameter.KEYWORD_ONLY:
                    kwonly_arg_types[p.name] = annotations.get(p.name, None) or Any
                elif p.kind in (_inspect.Parameter.VAR_POSITIONAL, _inspect.Parameter.VAR_KEYWORD):
                    raise TypeError("Variadic callables are not supported")

            result: Any = Callable[arg_types, return_type]
            if annotate_callable_kwargs:
                result = Annotated[result, CallableKwargsInfo(kwonly_arg_types)]
            return result
        except Exception:
            return Callable

    return type(value)


# TODO(egparedes): traversing a typing definition is an operation needed in several places
#   but it currently requires custom and cumbersome code due to the messy implementation details
#   in the standard library. Ideally, this code could be replaced by translating it once to a
#   custom "typing tree" data structure which could be then traversed in a generic way.
#



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import pytest


class TestFrozenList:
    def test_immutability(self):
        from gt4py.eve.type_definitions import FrozenList

        fl = FrozenList([0, 1, 2, 3, 4, 5])

        with pytest.raises(TypeError, match="object does not support item assignment"):
            fl[2] = -2

    def test_instance_check(self):
        from gt4py.eve.type_definitions import FrozenList

        assert isinstance(FrozenList([0, 1, 2, 3, 4, 5]), FrozenList)
        assert isinstance((), FrozenList)
        assert not isinstance([], FrozenList)


def test_sentinel():
    from gt4py.eve.type_definitions import NOTHING

    values = [0, 1, 2, NOTHING, 4, 6]

    assert values.index(NOTHING) == 3
    assert values[values.index(NOTHING)] is NOTHING



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import List

from gt4py.cartesian.gtc import oir
from gt4py.eve import NodeTranslator

from .utils import AccessCollector


class MaskStmtMerging(NodeTranslator):
    def _merge(self, stmts: List[oir.Stmt]) -> List[oir.Stmt]:
        merged = [self.visit(stmts[0])]
        for stmt in stmts[1:]:
            stmt = self.visit(stmt)
            if (
                isinstance(stmt, oir.MaskStmt)
                and isinstance(merged[-1], oir.MaskStmt)
                and stmt.mask == merged[-1].mask
                and not (
                    AccessCollector.apply(merged[-1].body).write_fields()
                    & AccessCollector.apply(stmt.mask, is_write=False).fields()
                )
            ):
                merged[-1] = oir.MaskStmt(mask=merged[-1].mask, body=merged[-1].body + stmt.body)
            else:
                merged.append(stmt)
        return merged

    def visit_HorizontalExecution(self, node: oir.HorizontalExecution) -> oir.HorizontalExecution:
        return oir.HorizontalExecution(
            body=self._merge(node.body),
            declarations=node.declarations,
            loc=node.loc,
        )

    # Stmt node types with lists of Stmts within them:

    def visit_MaskStmt(self, node: oir.MaskStmt) -> oir.MaskStmt:
        return oir.MaskStmt(mask=node.mask, body=self._merge(node.body), loc=node.loc)

    def visit_While(self, node: oir.While) -> oir.While:
        body_nodes = []
        for stmt in node.body:
            if isinstance(stmt, oir.MaskStmt) and node.cond == stmt.mask:
                body_nodes.extend(stmt.body)
            else:
                body_nodes.append(stmt)
        return oir.While(cond=self.visit(node.cond), body=self.visit(body_nodes), loc=node.loc)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numbers
from typing import Any, Optional, Sequence, Union

import numpy as np


try:
    import cupy as cp
except ImportError:
    cp = None

try:
    import dace
except ImportError:
    dace = None

if np.lib.NumpyVersion(np.__version__) >= "1.20.0":
    from numpy.typing import ArrayLike, DTypeLike
else:
    ArrayLike = Any
    DTypeLike = Any

from . import layout
from . import utils as storage_utils


def _error_on_invalid_backend(backend):
    if backend not in layout.REGISTRY:
        raise RuntimeError(f"Storage preset '{backend}' is not registered.")


def empty(
    shape: Sequence[int],
    dtype: DTypeLike = np.float64,
    *,
    backend: str,
    aligned_index: Optional[Sequence[int]] = None,
    dimensions: Optional[Sequence[str]] = None,
) -> Union[np.ndarray, "cp.ndarray"]:
    """Allocate an array of uninitialized (undefined) values with performance-optimal strides and alignment.

    Parameters
    ----------
        shape : `Sequence` of `int`
            The shape of the resulting `ndarray`
        dtype :  DTypeLike, optional
            The dtype of the resulting `ndarray`

    Keyword Arguments
    -----------------
        backend : `str`
            The target backend for which the allocation is optimized.
        aligned_index: `Sequence` of `int`, optional
            Indicate the index of the resulting array that most commonly corresponds to the origin of the compute
            domain. If not passed, it is aligned at the data origin.
        dimensions: `Sequence` of `str`, optional
            Indicate the semantic meaning of the dimensions in the provided array. Only used for determining optimal
            strides, the information is not stored.

    Returns
    -------
        NumPy or CuPy ndarray
            With uninitialized values, padded and aligned to provide optimal performance for the given `backend` and
            `aligned_index`

    Raises
    -------
        TypeError
            If arguments of an unexpected type are specified.
        ValueError
            If illegal or inconsistent arguments are specified.
    """
    _error_on_invalid_backend(backend)
    storage_info = layout.from_name(backend)
    assert storage_info is not None
    if storage_info["device"] == "gpu":
        allocate_f = storage_utils.allocate_gpu
    else:
        allocate_f = storage_utils.allocate_cpu

    aligned_index, shape, dtype, dimensions = storage_utils.normalize_storage_spec(
        aligned_index, shape, dtype, dimensions
    )

    _error_on_invalid_backend(backend)

    alignment = storage_info["alignment"]
    layout_map = storage_info["layout_map"](dimensions)

    dtype = np.dtype(dtype)
    _, res = allocate_f(shape, layout_map, dtype, alignment * dtype.itemsize, aligned_index)
    return res


def ones(
    shape: Sequence[int],
    dtype: DTypeLike = np.float64,
    *,
    backend: str,
    aligned_index: Optional[Sequence[int]] = None,
    dimensions: Optional[Sequence[str]] = None,
) -> Union[np.ndarray, "cp.ndarray"]:
    """Allocate an array with values initialized to 1.0 with performance-optimal strides and alignment.

    Parameters
    ----------
        shape : `Sequence` of `int`
            The shape of the resulting `ndarray`
        dtype :  DTypeLike, optional
            The dtype of the resulting `ndarray`

    Keyword Arguments
    -----------------
        backend : `str`
            The target backend for which the allocation is optimized.
        aligned_index: `Sequence` of `int`, optional
            Indicate the index of the resulting array that most commonly corresponds to the origin of the compute
            domain. If not passed, it is aligned at the data origin.
        dimensions: `Sequence` of `str`, optional
            Indicate the semantic meaning of the dimensions in the provided array. Only used for determining optimal
            strides, the information is not stored.

    Returns
    -------
        NumPy or CuPy ndarray
            Initialized to 1.0, padded and aligned to provide optimal performance for the given `backend` and
            `aligned_index`

    Raises
    -------
        TypeError
            If arguments of an unexpected type are specified.
        ValueError
            If illegal or inconsistent arguments are specified.
    """
    storage = empty(
        shape=shape,
        dtype=dtype,
        backend=backend,
        aligned_index=aligned_index,
        dimensions=dimensions,
    )
    storage[...] = storage.dtype.type(1)
    return storage


def full(
    shape: Sequence[int],
    fill_value: numbers.Number,
    dtype: DTypeLike = np.float64,
    *,
    backend: str,
    aligned_index: Optional[Sequence[int]] = None,
    dimensions: Optional[Sequence[str]] = None,
) -> Union[np.ndarray, "cp.ndarray"]:
    """Allocate an array with values initialized to `fill_value` with performance-optimal strides and alignment.

    Parameters
    ----------
        shape : `Sequence` of `int`
            The shape of the resulting `ndarray`
        fill_value: `numbers.Number`
            The value to which the array elements are initialized.
        dtype :  DTypeLike, optional
            The dtype of the resulting `ndarray`

    Keyword Arguments
    -----------------
        backend : `str`
            The target backend for which the allocation is optimized.
        aligned_index: `Sequence` of `int`, optional
            Indicate the index of the resulting array that most commonly corresponds to the origin of the compute
            domain. If not passed, it is aligned at the data origin.
        dimensions: `Sequence` of `str`, optional
            Indicate the semantic meaning of the dimensions in the provided array. Only used for determining optimal
            strides, the information is not stored.

    Returns
    -------
        NumPy or CuPy ndarray
            Initialized to `fill_value`, padded and aligned to provide optimal performance for the given `backend` and
            `aligned_index`

    Raises
    -------
        TypeError
            If arguments of an unexpected type are specified.
        ValueError
            If illegal or inconsistent arguments are specified.
    """
    storage = empty(
        shape=shape,
        dtype=dtype,
        backend=backend,
        aligned_index=aligned_index,
        dimensions=dimensions,
    )
    storage[...] = storage.dtype.type(fill_value)
    return storage


def zeros(
    shape: Sequence[int],
    dtype: DTypeLike = np.float64,
    *,
    backend: str,
    aligned_index: Optional[Sequence[int]] = None,
    dimensions: Optional[Sequence[str]] = None,
) -> Union[np.ndarray, "cp.ndarray"]:
    """Allocate an array with values initialized to 0.0 with performance-optimal strides and alignment.

    Parameters
    ----------
        shape : `Sequence` of `int`
            The shape of the resulting `ndarray`
        dtype :  DTypeLike, optional
            The dtype of the resulting `ndarray`

    Keyword Arguments
    -----------------
        backend : `str`
            The target backend for which the allocation is optimized.
        aligned_index: `Sequence` of `int`, optional
            Indicate the index of the resulting array that most commonly corresponds to the origin of the compute
            domain. If not passed, it is aligned at the data origin.
        dimensions: `Sequence` of `str`, optional
            Indicate the semantic meaning of the dimensions in the provided array. Only used for determining optimal
            strides, the information is not stored.

    Returns
    -------
        NumPy or CuPy ndarray
            Initialized to 0.0, padded and aligned to provide optimal performance for the given `backend` and
            `aligned_index`

    Raises
    -------
        TypeError
            If arguments of an unexpected type are specified.
        ValueError
            If illegal or inconsistent arguments are specified.
    """
    storage = empty(
        shape=shape,
        dtype=dtype,
        backend=backend,
        aligned_index=aligned_index,
        dimensions=dimensions,
    )
    storage[...] = storage.dtype.type(0)
    return storage


def from_array(
    data: ArrayLike,
    dtype: DTypeLike = np.float64,
    *,
    backend: str,
    aligned_index: Optional[Sequence[int]] = None,
    dimensions: Optional[Sequence[str]] = None,
) -> Union[np.ndarray, "cp.ndarray"]:
    """Allocate an array with values initialized to those of `data` with performance-optimal strides and alignment.

    This copies the values from `data` to the resulting buffer.

    Parameters
    ----------
        data : `ArrayLike`
            Source data to be copied, from which also the `shape` of the result is derived.
        dtype :  DTypeLike, optional
            The dtype of the resulting `ndarray`

    Keyword Arguments
    -----------------
        backend : `str`
            The target backend for which the allocation is optimized.
        aligned_index: `Sequence` of `int`, optional
            Indicate the index of the resulting array that most commonly corresponds to the origin of the compute
            domain. If not passed, it is aligned at the data origin.
        dimensions: `Sequence` of `str`, optional
            Indicate the semantic meaning of the dimensions in the provided array. Only used for determining optimal
            strides, the information is not stored.

    Returns
    -------
        NumPy or CuPy ndarray
            Copy of `data`, padded and aligned to provide optimal performance for the given `backend` and
            `aligned_index`

    Raises
    -------
        TypeError
            If arguments of an unexpected type are specified.
        ValueError
            If illegal or inconsistent arguments are specified.
    """
    is_cupy_array = cp is not None and isinstance(data, cp.ndarray)
    asarray = storage_utils.as_cupy if is_cupy_array else storage_utils.as_numpy
    shape = asarray(data).shape
    if dtype is None:
        dtype = asarray(data).dtype
    dtype = np.dtype(dtype)
    if dtype.shape:
        if dtype.shape and not shape[-dtype.ndim :] == dtype.shape:
            raise ValueError(f"Incompatible data shape {shape} with dtype of shape {dtype.shape}.")
        shape = shape[: -dtype.ndim]
    storage = empty(
        shape=shape,
        dtype=dtype,
        backend=backend,
        aligned_index=aligned_index,
        dimensions=dimensions,
    )

    if cp is not None and isinstance(storage, cp.ndarray):
        storage[...] = storage_utils.as_cupy(data)
    else:
        storage[...] = storage_utils.as_numpy(data)

    return storage


if dace is not None:

    def dace_descriptor(
        shape: Sequence[int],
        dtype: DTypeLike = np.float64,
        *,
        backend: str,
        aligned_index: Optional[Sequence[int]] = None,
        dimensions: Optional[Sequence[str]] = None,
    ) -> dace.data.Array:
        """Return a DaCe data descriptor which describes performance-optimal strides and alignment.

        Parameters
        ----------
            shape : `Sequence` of `int`
                The shape of the resulting data descriptor.
            dtype :  DTypeLike, optional
                The dtype encoded in the resulting data descriptor.

        Keyword Arguments
        -----------------
            backend : `str`
                The target backend for which the described allocation is optimized.
            aligned_index: `Sequence` of `int`, optional
                Indicate the index of the resulting array that most commonly corresponds to the origin of the compute
                domain. If not passed, it is aligned at the data origin.
            dimensions: `Sequence` of `str`, optional
                Indicate the semantic meaning of the dimensions in the data descriptor. Only used for determining
                optimal strides, the information is not stored.

        Returns
        -------
            DaCe data descriptor
                With strides that encode padding and aligned to provide optimal performance for the given `backend` and
                `aligned_index`

        Raises
        -------
            TypeError
                If arguments of an unexpected type are specified.
            ValueError
                If illegal or inconsistent arguments are specified.
        """
        aligned_index, shape, dtype, dimensions = storage_utils.normalize_storage_spec(
            aligned_index, shape, dtype, dimensions
        )
        itemsize = dtype.itemsize
        storage_info = layout.from_name(backend)
        assert storage_info is not None
        layout_map = storage_info["layout_map"](dimensions)

        order_idx = storage_utils.idx_from_order([i for i in layout_map if i is not None])
        padded_shape = storage_utils.compute_padded_shape(
            shape, storage_info["alignment"], order_idx
        )

        strides = storage_utils.strides_from_padded_shape(padded_shape, order_idx, itemsize)

        storage = (
            dace.StorageType.GPU_Global
            if storage_info["device"] == "gpu"
            else dace.StorageType.CPU_Heap
        )
        start_offset = int(np.array([aligned_index]) @ np.array([strides]).T) // itemsize

        total_size = int(int(np.array([shape]) @ np.array([strides]).T) // itemsize)

        start_offset = start_offset % storage_info["alignment"]
        return dace.data.Array(
            shape=shape,
            strides=[s // itemsize for s in strides],
            dtype=dace.typeclass(str(dtype)),
            storage=storage,
            total_size=total_size,
            start_offset=start_offset,
        )
